import os, re, subprocess, logging, timeit
import pandas as pd
import gmsh

from fiqus.data.DataFiQuSConductor import Conductor, SolutionParameters
from fiqus.data.RegionsModelFiQuS import RegionsModel
from fiqus.utils.Utils import GmshUtils, FilesAndFolders

from fiqus.pro_assemblers.ProAssembler import ASS_PRO

logger = logging.getLogger('FiQuS')

class Solve:
    def __init__(self, fdm, GetDP_path, geometry_folder, mesh_folder, verbose=True):
        self.fdm = fdm
        self.GetDP_path = GetDP_path

        self.solution_folder = os.path.join(os.getcwd())
        self.geometry_folder = geometry_folder
        self.mesh_folder = mesh_folder

        self.mesh_file = os.path.join(self.mesh_folder, f"{self.fdm.general.magnet_name}.msh")
        self.pro_file = os.path.join(self.solution_folder, f"{self.fdm.general.magnet_name}.pro")
        self.regions_file = os.path.join(mesh_folder, f"{self.fdm.general.magnet_name}.regions")
        
        self.verbose = verbose
        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=fdm.run.verbosity_Gmsh)

        self.ass_pro = ASS_PRO(os.path.join(self.solution_folder, self.fdm.general.magnet_name))
        self.regions_model = FilesAndFolders.read_data_from_yaml(self.regions_file, RegionsModel)
        self.material_properties_model = None

        self.ed = {} # excitation dictionary
        self.rohm = {} # rohm parameter dictionary
        self.rohf = {} # rohf parameter dictionary

        gmsh.option.setNumber("General.Terminal", verbose)

    def read_ro_parameters(self, inputs_folder_path):
        """
        Function for reading the CSV files containing the reduced order parameters for ROHF and ROHM.
        The function expects CSV files with an descriptive first header row (i.e. 'alphas,kappas,taus') and following rows with comma seperated parameter values for each cell starting from cell 1. 

        :param inputs_folder_path: The full path to the folder with input files
        """
        class reduced_order_params:
            " This structure is used to access the parameter dictionaries inside the pro template with a convenient syntax 'mp.rohf' or 'mp.rohm'."
            def __init__(self, rohf_dict=None, rohm_dict=None):
                self.rohf = rohf_dict
                self.rohm = rohm_dict

        if self.fdm.magnet.solve.rohm.enable and self.fdm.magnet.solve.rohm.parameter_csv_file:
            parameter_file = os.path.join(inputs_folder_path, self.fdm.magnet.solve.rohm.parameter_csv_file)
            df = pd.read_csv(parameter_file, delimiter=',', engine='python')
            df.index += 1
            logger.info(f'Using ROHM parameters from file: {parameter_file}:')
            logger.info(df)
            self.rohm = df.to_dict(orient='dict')

        if self.fdm.magnet.solve.rohf.enable and self.fdm.magnet.solve.rohf.parameter_csv_file:
            parameter_file = os.path.join(inputs_folder_path, self.fdm.magnet.solve.rohf.parameter_csv_file)
            df = pd.read_csv(parameter_file, delimiter=',', engine='python')
            df.index += 1
            logger.info(f'Using ROHF parameters from file: {parameter_file}:')
            logger.info(df)
            self.rohf = df.to_dict(orient='dict')

        # For the moment the reduced order parameters are treated as material properties mp structure.
        self.material_properties_model = reduced_order_params(rohf_dict=self.rohf, rohm_dict=self.rohm)

    def read_excitation(self, inputs_folder_path):
        """
        Function for reading a CSV file for the 'from_file' excitation case.

        :param inputs_folder_path: The full path to the folder with input files.
        :type inputs_folder_path: str
        """
        if self.fdm.magnet.solve.source_parameters.source_type == 'piecewise' and self.fdm.magnet.solve.source_parameters.piecewise.source_csv_file:
            input_file = os.path.join(inputs_folder_path, self.fdm.magnet.solve.source_parameters.piecewise.source_csv_file)
            logger.info(f'Using excitation from file: {input_file}')
            df = pd.read_csv(input_file, delimiter=',', engine='python')
            excitation_time = df['time'].to_numpy(dtype='float').tolist()
            self.ed['time'] = excitation_time
            excitation_value = df['value'].to_numpy(dtype='float').tolist()
            self.ed['value'] = excitation_value
                
    def assemble_pro(self):
        logger.info("Assembling .pro file")
        self.ass_pro.assemble_combined_pro(template = self.fdm.magnet.solve.pro_template, rm = self.regions_model, dm = self.fdm, ed=self.ed, mp=self.material_properties_model)

    def run_getdp(self, solve = True, postOperation = True, gui = False):

        command = ["-v2", "-verbose", "3", "-mat_mumps_icntl_14","100"]
        if solve: 
            command += ["-solve", "MagDyn"]
        command += ["-pos", "MagDyn"]

        if self.fdm.magnet.solve.general_parameters.noOfMPITasks:
            mpi_prefix = ["mpiexec", "-np", str(self.fdm.magnet.solve.general_parameters.noOfMPITasks)]
        else:
            mpi_prefix = []

        startTime = timeit.default_timer()

        getdpProcess = subprocess.Popen(mpi_prefix + [self.GetDP_path, self.pro_file, "-msh", self.mesh_file] + command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        with getdpProcess.stdout:
            for line in iter(getdpProcess.stdout.readline, b""):
                line = line.decode("utf-8").rstrip()
                line = line.split("\r")[-1]
                if not "Test" in line:
                    if line.startswith("Info"):
                        parsedLine = re.sub(r"Info\s+:\s+", "", line)
                        logger.info(parsedLine)
                    elif line.startswith("Warning"):
                        parsedLine = re.sub(r"Warning\s+:\s+", "", line)
                        logger.warning(parsedLine)
                    elif line.startswith("Error"):
                        parsedLine = re.sub(r"Error\s+:\s+", "", line)
                        logger.error(parsedLine)
                        logger.error("Solving HomogenizedConductor failed.")
                        # raise Exception(parsedLine)
                    elif re.match("##", line):
                        logger.critical(line)
                    else:
                        logger.info(line)
        
        simulation_time = timeit.default_timer()-startTime
        # Save simulation time:
        if solve:
            logger.info(f"Solving HomogenizedConductor has finished in {round(simulation_time, 3)} seconds.")
            with open(os.path.join(self.solution_folder, 'txt_files', 'simulation_time.txt'), 'w') as file:
                file.write(str(simulation_time))


        if gui and ((postOperation and not solve) or (solve and postOperation and self.fdm.magnet.postproc.generate_pos_files)):
            # gmsh.option.setNumber("Geometry.Volumes", 1)
            # gmsh.option.setNumber("Geometry.Surfaces", 1)
            # gmsh.option.setNumber("Geometry.Curves", 1)
            # gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in reversed(posFiles):
                gmsh.open(os.path.join(self.solution_folder, posFile))
            self.gu.launch_interactive_GUI()
        else:
            if gmsh.isInitialized():
                gmsh.clear()
                gmsh.finalize()

    def cleanup(self):
        """
            This funtion is used to remove .msh, .pre and .res files from the solution folder, as they may be large and not needed.
        """
        magnet_name = self.fdm.general.magnet_name
        cleanup = self.fdm.magnet.postproc.cleanup

        if cleanup.remove_res_file:
            res_file_path = os.path.join(self.solution_folder, f"{magnet_name}.res")
            if os.path.exists(res_file_path):
                os.remove(res_file_path)
                logger.info(f"Removed {magnet_name}.res")
        
        if cleanup.remove_pre_file:
            pre_file_path = os.path.join(self.solution_folder, f"{magnet_name}.pre")
            if os.path.exists(pre_file_path):
                os.remove(pre_file_path)
                logger.info(f"Removed {magnet_name}.pre")

        if cleanup.remove_msh_file:
            msh_file_path = os.path.join(self.mesh_folder, f"{magnet_name}.msh")
            if os.path.exists(msh_file_path):
                os.remove(msh_file_path)
                logger.info(f"Removed {magnet_name}.msh")


import os, json, gmsh

from fiqus.data import RegionsModelFiQuS
from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel
    
class Mesh:
    def __init__(self, fdm, verbose=True):
        """
        A base-class used to manage the mesh for HomogenizedConductor model.

        :ivar fdm: The fiqus data model for input parameters.
        :vartype fdm: dict
        """
        self.fdm = fdm
        self.mesh_folder = os.path.join(os.getcwd())
        self.geom_folder = os.path.dirname(self.mesh_folder)
        self.mesh_file = os.path.join(self.mesh_folder, f"{self.fdm.general.magnet_name}.msh")
        self.regions_file = os.path.join(self.mesh_folder, f"{self.fdm.general.magnet_name}.regions")
        self.vi_file = os.path.join(self.geom_folder, f'{self.fdm.general.magnet_name}.vi')
        self.verbose = verbose

        # dictionaries for physical groups
        self.dimTags_physical_surfaces = {}
        self.dimTags_physical_boundaries = {}
        self.dimTags_physical_cuts = {}
        self.dimTags_physical_points ={}

        # Read volume information file:
        with open(self.vi_file, "r") as f:
            self.dimTags = json.load(f)

        for key, value in self.dimTags.items():
            self.dimTags[key] = tuple(value) # dimTags contains all surfaces

        self.gu = GmshUtils(self.mesh_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=fdm.run.verbosity_Gmsh)

        gmsh.option.setNumber("General.Terminal", verbose)

    def generate_mesh(self):
        """ This function generates a mesh based on the volume information (vi) file created in the geometry step and the yaml input. """

        self.generate_physical_groups()

        # scale by domain size
        domain_size = self.fdm.magnet.geometry.air.radius
        # Mesh size field for cables:
        strand_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(strand_field, "SurfacesList", [value[1] for key, value in self.dimTags.items() if 'Cable' in key])
        gmsh.model.mesh.field.setNumber(strand_field, "VIn", self.fdm.magnet.mesh.cable_mesh_size_ratio * domain_size)
        # Mesh size field for excitation coils:
        coil_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(coil_field, "SurfacesList", [value[1] for key, value in self.dimTags.items() if 'Coil' in key])
        gmsh.model.mesh.field.setNumber(coil_field, "VIn", self.fdm.magnet.mesh.cable_mesh_size_ratio * domain_size)
        # Mesh size field for air:
        air_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(air_field, "SurfacesList", [self.dimTags['Air'][1]])
        gmsh.model.mesh.field.setNumber(air_field, "VIn", self.fdm.magnet.mesh.air_boundary_mesh_size_ratio * domain_size)

        total_meshSize_field = gmsh.model.mesh.field.add("Min")
        gmsh.model.mesh.field.setNumbers(total_meshSize_field, "FieldsList", [strand_field, coil_field, air_field])
        gmsh.model.mesh.field.setAsBackgroundMesh(total_meshSize_field)

        gmsh.option.setNumber("Mesh.MeshSizeFactor", self.fdm.magnet.mesh.scaling_global)
        gmsh.model.mesh.generate(2)


    def generate_physical_groups(self):
        """ This function generates the physical groups within the mesh based on the volume information (vi) file and stores their tags in dictionaries.

        :raises ValueError: For unknown volume names in the vi-file
        """
        gmsh.model.occ.synchronize()

        for key, value in self.dimTags.items():
            # surface
            surf_tag = gmsh.model.addPhysicalGroup(dim=value[0], tags=[value[1]], name=key)
            self.dimTags_physical_surfaces.update({str(key):(2, surf_tag)})
            # (outer) boundary curves
            if 'Air' in key:
                boundary_curves = gmsh.model.get_boundary([tuple(val) for val in self.dimTags.values()], combined=True)
            elif 'Cable' in key:
                boundary_curves = gmsh.model.get_boundary([tuple(value)]) 
            elif 'Coil' in key:
                boundary_curves = gmsh.model.get_boundary([tuple(value)]) 
            else:
                raise ValueError('Unknown volume in declaration in VI file.')
            bnd_tag = gmsh.model.addPhysicalGroup(dim=1, tags=[dimTag[1] for dimTag in boundary_curves], name=key+'Boundary' )
            self.dimTags_physical_boundaries.update({str(key+'Boundary' ):(1, bnd_tag)})
            # arbitrary boundary point
            point_tag = gmsh.model.addPhysicalGroup(dim=0, tags=[gmsh.model.get_boundary(boundary_curves, combined=False)[0][1]], name=key+'BoundaryPoint' )
            self.dimTags_physical_points.update({str(key+'BoundaryPoint' ):(1, point_tag)})


    def generate_cuts(self):
        """ This function generates and orients the domain cuts for the cable and coil regions through the gmsh internal homology module."""

        dimTags_physical_cable_boundaries = {k: v for k, v in self.dimTags_physical_boundaries.items() if 'Cable' in k}
        dimTags_physical_excitation_coil_boundaries = {k: v for k, v in self.dimTags_physical_boundaries.items() if 'Coil' in k}

        tags_physical_cable_surfaces = [v[1] for k, v in self.dimTags_physical_surfaces.items() if 'Cable' in k]
        tags_physical_excitation_coil_surfaces = [v[1] for k, v in self.dimTags_physical_surfaces.items() if 'Coil' in k]
        
        # print(dimTags_physical_cable_boundaries)
        # print(dimTags_physical_excitation_coil_boundaries)
        # print(tags_physical_cable_surfaces)
        # print(tags_physical_excitation_coil_surfaces)

        # cohomology cuts for all cables
        for _, value in dimTags_physical_cable_boundaries.items():
            gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[value[1]], dims=[1])
        for _, value in dimTags_physical_excitation_coil_boundaries.items():
            gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[value[1]], dims=[1])

        gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[self.dimTags_physical_surfaces['Air'][1]]+tags_physical_excitation_coil_surfaces, dims=[1])
        for coil_surface in tags_physical_excitation_coil_surfaces:
            tags_physical_excitation_other_coil_surfaces = [other_coil for other_coil in tags_physical_excitation_coil_surfaces if other_coil != coil_surface]
            gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[self.dimTags_physical_surfaces['Air'][1]]+tags_physical_cable_surfaces+tags_physical_excitation_other_coil_surfaces, dims=[1])



        dimTags_homology = gmsh.model.mesh.computeHomology()
        # print(dimTags_homology)
        dimTags_bnds = dimTags_homology[:int(len(dimTags_homology)/2)] # first half are dimTags are the cut boundaries
        dimTags_cuts = dimTags_homology[int(len(dimTags_homology)/2):] # second half are the actual cut edges
        gmsh.model.mesh.clearHomologyRequests()
        
        # post process homology cuts
        bnd_tags = []
        cut_tags = []
        for i in range(len(dimTags_physical_cable_boundaries)):
            self.dimTags_physical_cuts.update({'Cable'+str(i+1)+'Cut':(1, dimTags_cuts[-1][1]+(i+1))}) # +1 tag shift for post processed cuts
            bnd_tags.append(dimTags_bnds[i][1])
            cut_tags.append(dimTags_cuts[i][1])
        for i in range(len(dimTags_physical_excitation_coil_boundaries)):
            self.dimTags_physical_cuts.update({'Coil'+str(i+1)+'Cut':(1, dimTags_cuts[-1][1]+(len(dimTags_physical_cable_boundaries)+i+1))}) # +1 tag shift for post processed cuts
            bnd_tags.append(dimTags_bnds[len(dimTags_physical_cable_boundaries)+i][1])
            cut_tags.append(dimTags_cuts[len(dimTags_physical_cable_boundaries)+i][1])
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", ','.join(map(str,bnd_tags)))
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", ','.join(map(str,cut_tags)))
        gmsh.plugin.run("HomologyPostProcessing")


    def generate_regions_file(self):
        """
        Generates a .regions file for the GetDP solver, containing all necessary information about the model.
        The regions model contains data about physical surfaces, boundaries, points and cuts and is stored in the mesh folder.

        :raises ValueError: For unknown volumes in the vi-file
        """
        regions_model = RegionsModel()

        # The region Cables will include the homogenized surfaces
        regions_model.powered['Cables'] = RegionsModelFiQuS.Powered()
        regions_model.powered['Cables'].vol.names = []
        regions_model.powered['Cables'].vol.numbers = []
        regions_model.powered['Cables'].surf.names = []
        regions_model.powered['Cables'].surf.numbers = [] 
        regions_model.powered['Cables'].curve.names = []
        regions_model.powered['Cables'].curve.numbers = [] 
        regions_model.powered['Cables'].cochain.names = []
        regions_model.powered['Cables'].cochain.numbers = [] 
        # Excitation coil regions
        regions_model.powered['ExcitationCoils'] = RegionsModelFiQuS.Powered()
        regions_model.powered['ExcitationCoils'].vol.names = []
        regions_model.powered['ExcitationCoils'].vol.numbers = []
        regions_model.powered['ExcitationCoils'].surf.names = []
        regions_model.powered['ExcitationCoils'].surf.numbers = [] 
        regions_model.powered['ExcitationCoils'].curve.names = []
        regions_model.powered['ExcitationCoils'].curve.numbers = [] 
        regions_model.powered['ExcitationCoils'].cochain.names = []
        regions_model.powered['ExcitationCoils'].cochain.numbers = [] 
       
        gmsh.model.occ.synchronize()
        for name in self.dimTags.keys():
            cut_name = name+'Cut'
            boundary_name = name+'Boundary'
            point_name = boundary_name+'Point'
            if 'Air' in name:
                regions_model.air.vol.number = self.dimTags_physical_surfaces[name][1]
                regions_model.air.vol.name = "Air"
                regions_model.air.surf.number = self.dimTags_physical_boundaries[boundary_name][1]
                regions_model.air.surf.name = boundary_name
                regions_model.air.point.names = [point_name]
                regions_model.air.point.numbers = [self.dimTags_physical_points[point_name][1]]
            elif 'Cable' in name:
                regions_model.powered['Cables'].vol.names.append(name)
                regions_model.powered['Cables'].vol.numbers.append(self.dimTags_physical_surfaces[name][1])
                regions_model.powered['Cables'].surf.names.append(boundary_name)
                regions_model.powered['Cables'].surf.numbers.append(self.dimTags_physical_boundaries[boundary_name][1])
                regions_model.powered['Cables'].curve.names.append(point_name)
                regions_model.powered['Cables'].curve.numbers.append(self.dimTags_physical_points[point_name][1])
                regions_model.powered['Cables'].cochain.names.append(cut_name)
                regions_model.powered['Cables'].cochain.numbers.append(self.dimTags_physical_cuts[cut_name][1])
            elif 'Coil' in name:
                regions_model.powered['ExcitationCoils'].vol.names.append(name)
                regions_model.powered['ExcitationCoils'].vol.numbers.append(self.dimTags_physical_surfaces[name][1])
                regions_model.powered['ExcitationCoils'].surf.names.append(boundary_name)
                regions_model.powered['ExcitationCoils'].surf.numbers.append(self.dimTags_physical_boundaries[boundary_name][1])
                regions_model.powered['ExcitationCoils'].curve.names.append(point_name)
                regions_model.powered['ExcitationCoils'].curve.numbers.append(self.dimTags_physical_points[point_name][1])
                regions_model.powered['ExcitationCoils'].cochain.names.append(cut_name)
                regions_model.powered['ExcitationCoils'].cochain.numbers.append(self.dimTags_physical_cuts[cut_name][1])
            else:
                raise ValueError('Unknown physical region')

        FilesAndFolders.write_data_to_yaml(self.regions_file, regions_model.model_dump())

    def save_mesh(self, gui: bool = False):
        """ Saves the mesh to a .msh file. If gui is True, the mesh is also loaded in the gmsh GUI. """
        
        gmsh.write(self.mesh_file)
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            if gmsh.isInitialized():
                gmsh.clear()
                gmsh.finalize()

    def load_mesh(self, gui : bool = False):
        """ Loads a previously generated mesh. """
        
        gmsh.clear()
        gmsh.open(self.mesh_file)

        if gui:
            self.gu.launch_interactive_GUI()




from pydantic import BaseModel, Field
from typing import List, Dict, Literal, Union, Optional, Tuple


class MultipoleGeoElement(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    lines: Optional[int] = Field(
        default=3,
        description="It specifies the number of Gaussian points for lines.",
    )
    triangles: Optional[Literal[1, 3, 4, 6, 7, 12, 13, 16]] = Field(
        default=3,
        description="It specifies the number of Gaussian points for triangles.",
    )
    quadrangles: Optional[Literal[1, 3, 4, 7]] = Field(
        default=4,
        description="It specifies the number of Gaussian points for quadrangles.",
    )


class MultipoleSolveConvectionBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    heat_transfer_coefficient: Optional[Union[float, str]] = Field(
        default=None,
        description="It specifies the value or function name of the heat transfer coefficient for this boundary condition.",
    )


class MultipoleSolveHeatFluxBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    const_heat_flux: Optional[float] = Field(
        default=None,
        description="It specifies the value of the heat flux for this boundary condition.",
    )
    # function_heat_flux: Optional[str] = None


class MultipoleSolveTemperatureBoundaryCondition(BaseModel):
    """
    Level 5: Class for FiQuS Multipole
    """
    boundaries: Optional[List[str]] = Field(
        default=[],
        description="It specifies the list of boundaries where the condition is applied."
                    "Each boundary is identified by a string of the form <half-turn/wedge reference number><side>,"
                    "where the accepted sides are i, o, l, h which correspond respectively to inner, outer, lower (angle), higher (angle): e.g., 1o",
    )
    const_temperature: Optional[float] = Field(
        default=None,
        description="It specifies the value of the temperature for this boundary condition.",
    )
    # function_temperature: Optional[str] = None


class MultipoleSolveQuenchInitiation(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    turns: Optional[List[int]] = Field(
        default=[],
        description="It specifies the list of reference numbers of half-turns whose critical currents are set to zero.",
    )
    t_trigger: Optional[List[float]] = Field(
        default=[],
        description="It specifies the list of time instants at which the critical current is set to zero.",
    )


class MultipoleSolveBoundaryConditionsThermal(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    temperature: Optional[Dict[str, MultipoleSolveTemperatureBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Dirichlet boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )
    heat_flux: Optional[Dict[str, MultipoleSolveHeatFluxBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Neumann boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )
    cooling: Optional[Dict[str, MultipoleSolveConvectionBoundaryCondition]] = Field(
        default={},
        description="This dictionary contains the information about the Robin boundary conditions."
                    "The keys are chosen names for each boundary condition.",
    )

class MultipoleSolveTransient_parent(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    initial_time: Optional[float] = Field(
        default=0.,
        description="It specifies the initial time of the simulation.",
    )
    final_time: Optional[float] = Field(
        default=0.0,
        description="It specifies the final time of the simulation.",
    )
    initial_time_step: Optional[float] = Field(
        default=1E-10,
        description="It specifies the initial time step used at the beginning of the transient simulation.",
    )
    min_time_step: Optional[float] = Field(
        default=1E-12,
        description="It specifies the minimum possible value of the time step.",
    )
    max_time_step: Optional[float] = Field(
        default=10,
        description="It specifies the maximum possible value of the time step.",
    )
    breakpoints: Optional[List[float]] = Field(
        default=[],
        description="It forces the transient simulation to hit the time instants contained in this list.",
    )
    integration_method: Optional[Union[None, Literal[
        "Euler", "Gear_2", "Gear_3", "Gear_4", "Gear_5", "Gear_6"
    ]]] = Field(
        default="Euler",
        title="Integration Method",
        description="It specifies the type of integration method to be used.",
    )
    rel_tol_time: Optional[float] = Field(
        default=1E-4,
        description="It specifies the relative tolerance.",
    )
    abs_tol_time: Optional[float] = Field(
        default=1e-4,
        description="It specifies the absolute tolerance.",
    )
    norm_type: Optional[Literal["L1Norm", "MeanL1Norm", "L2Norm", "MeanL2Norm", "LinfNorm"]] = Field(
        default='LinfNorm',
        description="It specifies the type of norm to be calculated for convergence assessment.",
    )

class MultipoleSolveTransientElectromagnetics(MultipoleSolveTransient_parent):
    """
    Level 4: Class for FiQuS Multipole
    """
    T_sim: Optional[float] = Field(
        default=1.9,
        description="It specifies the temperature used to calculate the resistivity of the superconductor during the transient sim.",
    )

class MultipleSolveCollarHeCooling(BaseModel):
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the helium cooling is enabled or not (adiabatic conditions).",
    )
    which: Optional[Union[Literal['all'], List]] = Field(
        default='all',
        description="It specifies the boundaries where the collar cooling is applied. If 'all', it applies to all boundaries. If a list, it applies to the specified boundaries numbered counter-clockwise."
    )
    heat_transfer_coefficient: Optional[Union[float, str]] = Field(
        default= 'CFUN_hHe_T_THe',
        description="It specifies the value or name of the function of the constant heat transfer coefficient.",
    )
    ref_temperature: Optional[float] = Field(
        default = 0.0,
        description="It specifies the reference temperature for the collar cooling. If not specified, it takes the value of the initial temperature.",
    )
    move_cooling_holes: Optional[Union[str, int, List[List[float]]]] = Field(
        default=None,
        description= "It specifies if and how cooling holes are to be moved. Either choose '1' or '2' for predefined positions or a list [[dx,dy], [dx2,dy2]].. to shift each hole manually"
    )

class MultipoleSolveHeCooling(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the helium cooling is enabled or not (adiabatic conditions).",
    )
    sides: Optional[Literal["external", "inner", "outer", "inner_outer"]] = Field(
        default="outer",
        description="It specifies the general grouping of the boundaries where to apply cooling:"
                    "'external': all external boundaries; 'inner': only inner boundaries; 'outer': only outer boundaries; 'inner_outer': inner and outer boundaries.",
    )
    heat_transfer_coefficient: Optional[Union[float, str]] = Field(
        default=0.0,
        description="It specifies the value or name of the function of the constant heat transfer coefficient.",
    )


class MultipoleSolveNonLinearSolver(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    rel_tolerance: Optional[float] = Field(
        default=1E-4,
        description="It specifies the relative tolerance.",
    )
    abs_tolerance: Optional[float] = Field(
        default=0.1,
        description="It specifies the absolute tolerance.",
    )
    relaxation_factor: Optional[float] = Field(
        default=0.7,
        description="It specifies the relaxation factor.",
    )
    max_iterations: Optional[int] = Field(
        default=20,
        description="It specifies the maximum number of iterations if no convergence is reached.",
    )
    norm_type: Literal["L1Norm", "MeanL1Norm", "L2Norm", "MeanL2Norm", "LinfNorm"] = Field(
        default='LinfNorm',
        description="It specifies the type of norm to be calculated for convergence assessment.",
    )
    integration_method: Union[None, Literal[
       "Euler", "Gear_2", "Gear_3", "Gear_4", "Gear_5", "Gear_6"
    ]] = Field(
        default="Euler",
        title="Integration Method",
        description="It specifies the type of integration method to be used.",
    )

class MultipoleSolveTransientThermal(MultipoleSolveTransient_parent):
    """
    Level 4: Class for FiQuS Multipole
    """
    stop_temperature: Optional[float] = Field(
        default=300,
        description="If one half turn reaches this temperature, the simulation is stopped.",
    )

class MultipoleSolveTransientCoupled(MultipoleSolveTransient_parent):
    """
    Level 4: Class for FiQuS Multipole
    """
    rel_tol_time: Optional[List[float]] = Field(
        default=[1E-4,1E-4],
        description="It specifies the relative tolerance.",
    )
    abs_tol_time: Optional[List[float]] = Field(
        default=[1e-4,1e-4],
        description="It specifies the absolute tolerance.",
    )
    norm_type: List[Literal["L1Norm", "MeanL1Norm", "L2Norm", "MeanL2Norm", "LinfNorm"]] = Field(
        default=['LinfNorm','LinfNorm'],
        description="It specifies the type of norm to be calculated for convergence assessment.",
    )
    stop_temperature: Optional[float] = Field(
        default=300,
        description="If one half turn reaches this temperature, the simulation is stopped.",
    )
    seq_NL: Optional[bool] = Field(
    default=True,
    description="The non-linear solver is sequential Mag->Thermal, or its fully coupled.",
    )

class MultipoleSolveInsulationBlockToBlock(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    It contains the information about the materials and thicknesses of the inner insulation regions (between blocks) modeled via thin-shell approximation.
    """
    material: Optional[str] = Field(
        default=None,
        description="It specifies the default material of the insulation regions between the blocks insulation regions.",
    )
    # the order of blocks should be: [inner, outer] for mid-layer couples or [lower, higher] for mid-pole and mid-winding couples
    blocks_connection_overwrite: List[Tuple[str, str]] = Field(
        default=[],
        description="It specifies the blocks couples adjacent to the insulation region."
                    "The blocks must be ordered from inner to outer block for mid-layer insulation regions and from lower to higher angle block for mid-pole and mid-winding insulation regions.",
    )
    materials_overwrite: Optional[List[List[str]]] = Field(
        default=[],
        description="It specifies the list of materials making up the layered insulation region to be placed between the specified blocks."
                    "The materials must be ordered from inner to outer layers and lower to higher angle layers.",
    )
    thicknesses_overwrite: Optional[List[List[Optional[float]]]] = Field(
        default=[],
        description="It specifies the list of thicknesses of the specified insulation layers. The order must match the one of the materials list.",
    )


class MultipoleSolveInsulationExterior(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    It contains the information about the materials and thicknesses of the outer insulation regions (exterior boundaries) modeled via thin-shell approximation.
    """
    blocks: Optional[List[str]] = Field(
        default=[],
        description="It specifies the reference numbers of the blocks adjacent to the exterior insulation regions to modify.",
    )
    materials_append: Optional[List[List[str]]] = Field(
        default=[],
        description="It specifies the list of materials making up the layered insulation region to be appended to the block insulation."
                    "The materials must be ordered from the block outward.",
    )
    thicknesses_append: Optional[List[List[float]]] = Field(
        default=[],
        description="It specifies the list of thicknesses of the specified insulation layers. The order must match the one of the materials list.",
    )

class MultipoleSolveSpecificMaterial(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    material: Optional[str] = Field(
        default=None,
        description="It specifies the material of the region.",
    )
    RRR: Optional[float] = Field(
        default=None,
        description="It specifies the RRR of the region.",
    )
    T_ref_RRR_high: Optional[float] = Field(
        default=None,
        description="It specifies the reference temperature associated with the RRR.",
    )
    transient_effects_enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the transient effects are enabled or not.",
    )
    # rel_magnetic_permeability: Optional[float] = Field(
    #     default = 1.0,
    #     description = 'It specifies the material relative magnetic permeability against vacuum for EM calculations'
    # )


class MultipoleSolveInsulationCollar(BaseModel):
    material: Optional[str] = Field(
        default=None,
        description="It specifies the default material of the insulation regions between collar and outer insulation.",
    )

class MultipoleSolveInsulationTSA(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    block_to_block: MultipoleSolveInsulationBlockToBlock = Field(
        default=MultipoleSolveInsulationBlockToBlock(),
        description="This dictionary contains the information about the materials and thicknesses of the inner insulation regions (between blocks) modeled via thin-shell approximation.",
    )
    exterior: Optional[MultipoleSolveInsulationExterior] = Field(
        default=MultipoleSolveInsulationExterior(),
        description="This dictionary contains the information about the materials and thicknesses of the outer insulation regions (exterior boundaries) modeled via thin-shell approximation.",
    )
    between_collar: Optional[MultipoleSolveInsulationBlockToBlock] = Field(
        default=MultipoleSolveInsulationCollar(),
        description="This dictionary contains the information about the materials and thicknesses of the insulation regions between the collar and the outer insulation regions for thin-shell approximation.",
    )

class MultipoleSolve_parent(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    non_linear_solver: MultipoleSolveNonLinearSolver = Field(
        default=MultipoleSolveNonLinearSolver(),
        description="This dictionary contains the information about the parameters for the non-linear solver.",
    )


class MultipoleSolveThermal(MultipoleSolve_parent):
    """
    Level 3: Class for FiQuS Multipole
    """
    solve_type: Optional[Literal[None, "transient"]] = Field(
        default=None,
        description="It determines whether the thermal transient problem is solved ('transient') or not ('null').",
    )
    insulation_TSA: Optional[MultipoleSolveInsulationTSA] = Field(
        default=MultipoleSolveInsulationTSA(),
        description="This dictionary contains the information about the materials and thicknesses of the insulation regions modeled via thin-shell approximation.",
    )
    He_cooling: MultipoleSolveHeCooling = Field(
        default=MultipoleSolveHeCooling(),
        description="This dictionary contains the information about the Robin boundary condition for generic groups of boundaries.",
    )
    collar_cooling: MultipleSolveCollarHeCooling = Field(
        default=MultipleSolveCollarHeCooling(),
        description="This dictionary contains the information about the cooling for the collar region.",
    )
    overwrite_boundary_conditions: Optional[MultipoleSolveBoundaryConditionsThermal] = Field(
        default=MultipoleSolveBoundaryConditionsThermal(),
        description="This dictionary contains the information about boundary conditions for explicitly specified boundaries.",
    )
    time_stepping: MultipoleSolveTransientThermal = Field(
        default=MultipoleSolveTransientThermal(),
        description="This dictionary contains the information about the parameters for the transient solver.",
    )
    jc_degradation_to_zero: Optional[MultipoleSolveQuenchInitiation] = Field(
        default=MultipoleSolveQuenchInitiation(),
        description="This dictionary contains the information about half turns with zero critical current.",
    )
    init_temperature: Optional[float] = Field(
        default=1.9,
        description="It specifies the initial temperature of the simulation.",
    )
    enforce_init_temperature_as_minimum: Optional[bool] = Field(
        default=False,
        description="It determines whether the initial temperature is enforced as the minimum temperature of the simulation.",
    )

class MultipoleSolveElectromagnetics(MultipoleSolve_parent):
    """
    Level 3: Class for FiQuS Multipole
    """
    solve_type: Optional[Literal[None, "stationary","transient"]] = Field(
        default=None,
        description="It determines whether the magneto-static problem is solved ('stationary') or not ('null').",
    )
    time_stepping: Optional[MultipoleSolveTransientElectromagnetics] = Field(
        default=MultipoleSolveTransientElectromagnetics(),
        description="This dictionary contains the information about the parameters for the transient solver.",
    )


class MultipoleMeshThinShellApproximationParameters(BaseModel):
    """
    Level 4: Class for FiQuS Multipole
    """
    minimum_discretizations: Optional[int] = Field(
        default=1,
        description="It specifies the number of minimum spacial discretizations across a thin-shell.",
    )
    global_size_QH: Optional[float] = Field(
        default=1e-4,
        description="The thickness of the quench heater region is divided by this parameter to determine the number of spacial discretizations across the thin-shell.",
    )
    minimum_discretizations_QH: Optional[int] = Field(
        default=1,
        description="It specifies the number of minimum spacial discretizations across a thin-shell.",
    )
    global_size_COL: Optional[float] = Field(
        default=1e-4,
        description="The thickness of the region between ht and collar is divided by this parameter to determine the number of spacial discretizations across the thin-shell.",
    )
    minimum_discretizations_COL: Optional[int] = Field(
        default=1,
        description="It specifies the number of minimum spacial discretizations across a thin-shell.",
    )
    scale_factor_radial: Optional[float] = Field(
        default=-1.0,
        description="Scaling factor for radially directed thin-shells (e.g. halfturns to collar). Set to -1.0 to use default scaling. Wedge scalings are always ignored.",
    )
    scale_factor_azimuthal: Optional[float] = Field(
        default=-1.0,
        description="Scaling factor for azimuthally directed thin-shells (e.g. halfturns to pole). Set to -1.0 to use default scaling. Wedge scalings are always ignored.",
    )

class MultipoleMeshThreshold(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the gmsh Field is enabled or not.",
    )
    SizeMin: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshSizeMin.",
    )
    SizeMax: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshSizeMax.",
    )
    DistMin: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshDistMin.",
    )
    DistMax: Optional[float] = Field(
        default=None,
        description="It sets gmsh Mesh.MeshDistMax.",
    )

class MultipoleMeshThresholdCollar(MultipoleMeshThreshold):
    """
    Level 3: Class for FiQuS Multipole
    """
    Enforce_TSA_mapping: Optional[bool] = Field(
        default=False,
        description="Enfocres matching nodes for the TSA layer. Uses SizeMin to determine the size of the nodes.", # only for the collar layer
    )

class MultipoleMeshTransfinite(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    enabled_for: Literal[None, "curves", "curves_and_surfaces"] = Field(
        default=None,
        description="It determines on what entities the transfinite algorithm is applied.",
    )
    curve_target_size_height: Optional[float] = Field(
        default=1.0,
        description="The height of the region (short side) is divided by this parameter to determine the number of elements to apply via transfinite curves.",
    )
    curve_target_size_width: Optional[float] = Field(
        default=1.0,
        description="The width of the region (long side) is divided by this parameter to determine the number of elements to apply via transfinite curves.",
    )
class MultipoleMeshTransfiniteOrField(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    transfinite: MultipoleMeshTransfinite = Field(
        default=MultipoleMeshTransfinite(),
        description="This dictionary contains the mesh information for transfinite curves.",
    )
    field: MultipoleMeshThreshold = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information.",
    )

class MultipolePostProc_parent(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    output_time_steps_pos: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    output_time_steps_txt: Optional[Union[bool, int]] = Field(
        default=True,
        description="It determines whether the solution for the .txt file is saved for all time steps (True), none (False), or equidistant time steps (int).",
    )
    save_pos_at_the_end: Optional[bool] = Field(
        default=True,
        description="It determines whether the solution for the .pos file is saved at the end of the simulation or during run time.",
    )
    save_txt_at_the_end: Optional[bool] = Field(
        default=False,
        description="It determines whether the solution for the .txt file is saved at the end of the simulation or during run time.",
    )

    plot_all: Optional[Union[bool, None]] = Field(
        default=False,
        description="It determines whether the figures are generated and shown (true), generated only (null), or not generated (false). Useful for tests.",
    )

class MultipolePostProcThermal(MultipolePostProc_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    take_average_conductor_temperature: Optional[bool] = Field(
        default=True,
        description="It determines whether the output files are based on the average conductor temperature or not (map2d).",
    )
    variables: Optional[List[Literal["T", "jOverJc", "rho", "az_thermal", "ac_loss"]]] = Field(
        default=["T"],
        description="It specifies the physical quantity to be output.",
    )
    volumes: Optional[List[
        Literal["omega", "powered", "induced", "iron", "conducting", "insulator"]]] = Field(
        default=["powered"],
        description="It specifies the regions associated with the physical quantity to be output.",
    )


class MultipolePostProcElectromagnetics(MultipolePostProc_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    compare_to_ROXIE: Optional[str] = Field(
        default=None,
        description="It contains the absolute path to a reference ROXIE map2d file. If provided, comparative plots with respect to the reference are generated.",
    )
    variables: Optional[List[Literal["a", "az", "b", "h", "js","jOverJc", "sigma_collar","is"]]] = Field(
        default=[],
        description="It specifies the physical quantity to be output.",
    )
    volumes: Optional[List[
        Literal["omega", "powered", "induced", "air", "air_far_field", "iron", "conducting", "insulator"]]] = Field(
        default=[],
        description="It specifies the regions associated with the physical quantity to be output.",
    )

class CCPostProc(BaseModel):
    variables_I: Optional[List[Literal["I_PC","I_1","I_2","I_cpc","I_crowbar","I_3","I_c_r","I_EE","I_c","I_s","I_C",
                                       "I_EE_n","I_c_n","I_s_n","I_QH","I_EQ","I_ESC",
                                       "I_A","I_B","I_C",
                                       "I_EQ",
                                       "I_ESC","I_ESC_Diode","I_ESC_C"]]] = Field(
        default=[],
        description="Currents from the circuit that will be exported as csv",
    )
    variables_U: Optional[List[Literal["PS_currentsource","PS_R_1","PS_L_1","PS_C","PS_R_3","PS_L_3","PS_R_2","PS_L_2","PS_R_crowbar","PS_Ud_crowbar","PS_L_crowbar","PS_R_c_r","PS_Ud_c_r","PS_L_c_r",
                                          "circ_R_circuit",
                                          "EE_L","EE_V_EE","EE_Ud_snubber","EE_C","EE_R_c","EE_L_c","EE_Ud_switch","EE_R_s","EE_L_s","EE_L_n","EE_V_EE_n","EE_Ud_snubber_n","EE_C_n","EE_R_c_n","EE_L_c_n","EE_Ud_switch_n","EE_R_s_n","EE_L_s_n","EE_R_switch","EE_R_switch_n",
                                          "CLIQ_R","CLIQ_L","CLIQ_C",
                                          "ECLIQ_currentsource","ECLIQ_L_leads","ECLIQ_R_leads",
                                          "ESC_C1","ESC_C2","ESC_R_leads","ESC_R_unit","ESC_L","ESC_L_Diode","ESC_Ud_Diode"]]] = Field(
        default=[],
        description="Voltages from the circuit that will be exported as csv",
    )
    assemble_veusz: Optional[bool] = Field(
        default=False,
        description="It determines whether the post-processing data is assembled in a veusz file.",
    )

class MultipolePostProc(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    electromagnetics: MultipolePostProcElectromagnetics = Field(
        default=MultipolePostProcElectromagnetics(),
        description="This dictionary contains the post-processing information for the electromagnetic solution.",
    )
    thermal: MultipolePostProcThermal = Field(
        default=MultipolePostProcThermal(),
        description="This dictionary contains the post-processing information for the thermal solution.",
    )
    circuit_coupling: CCPostProc = Field(
        default= CCPostProc(),
        description="This dictionary contains the post-processing information for the circuit variables calculated in the solution.",
    )

class MultipoleSolveCoilWindingsElectricalOrder(BaseModel):
    """
    Level 2: Class for the order of the electrical pairs
    """
    group_together: Optional[List[List[int]]] = []  # elPairs_GroupTogether
    reversed: Optional[List[int]] = []  # elPairs_RevElOrder
    overwrite_electrical_order: Optional[List[int]] = []

class MultipoleSolveCoilWindings(BaseModel):
    """
        Level 1: Class for winding information
    """
    conductor_to_group: Optional[List[int]] = []  # This key assigns to each group a conductor of one of the types defined with Conductor.name
    group_to_coil_section: Optional[List[int]] = []  # This key assigns groups of half-turns to coil sections
    polarities_in_group: Optional[List[int]] = []  # This key assigns the polarity of the current in each group #
    half_turn_length: Optional[List[float]] = []
    electrical_pairs: Optional[MultipoleSolveCoilWindingsElectricalOrder] = MultipoleSolveCoilWindingsElectricalOrder()  # Variables used to calculate half-turn electrical order
    # Homogenized Multipole
class HomogenizedConductorFormulationparametersROHM(BaseModel):
    """
    Level 4: Class for finite element formulation parameters
    """
    enabled: Optional[bool] = Field(
        default=False,
        description='Use ROHM to homogenize the magnetization hysteresis in the cables.'
    )
    parameter_csv_file: Optional[str] = Field(
        default=None,
        description='Name of the csv file containing the ROHM parameters within the inputs folder with expected row structure: [alpha,kappa,chi,gamma,lambda].'
    )
    gather_cell_systems: Optional[bool] = Field(
        default = False,
        description = 'when true, it generates a single system to solve the ROHM cells instead of one system per cell to decrease generation time.'
    )
    weight_scaling: Optional[float] = Field(
        default=1.0,
        description='Downscaling factor (s<1.0) which is applied to all weights except the first, which is scaled up to compensate.'
    )
    tau_scaling: Optional[float] = Field(
        default=1.0,
        description='Scaling factor which is applied uniformly to all coupling time constants.'
    )

class HomogenizedConductorFormulationparametersROHF(BaseModel):
    """
    Level 4: Class for finite element formulation parameters
    """
    enabled: Optional[bool] = Field(
        default=False,
        description='Use ROHF to homogenize the internal flux hysteresis in the cables.'
    )
    parameter_csv_file: Optional[str] = Field(
        default=None,
        description='Name of the csv file containing the ROHF parameters within the inputs folder with expected row structure: [alpha,kappa,tau].'
    )
    gather_cell_systems: Optional[bool] = Field(
        default = False,
        description = 'when true, it generates a single system to solve the ROHF cells instead of one system per cell to decrease generation time.'
    )
class HomogenizedConductorRunType(BaseModel):
    """
    Level 4: Class for runtype parameters
    """
    mode: Optional[Literal["ramp","isothermal_ramp","quench"]] = Field(
        default="ramp",
        description= "Type of simulation to run with homogenized conductors (ramp - real cooling conditions, isothermal_ramp - unlimited cooling, quench - non-zero initial conditions)"
    )
    ramp_file: Optional[str] = Field(
        default=None,
        description='Name of the ramp model from which to start the simulation'
    )
class HomogenizedConductor(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    enabled: Optional[bool] = Field(
        default=False,
        description="It determines whether the homogenized conductor model is enabled or not."
    )
    run_type: HomogenizedConductorRunType = Field(
        default=HomogenizedConductorRunType(),
        description= "Type of simulation to run with homogenized conductors (ramp - real cooling conditions, isothermal_ramp - unlimited cooling, quench - non-zero initial conditions)"
    )
    rohm: HomogenizedConductorFormulationparametersROHM = Field(
        default=HomogenizedConductorFormulationparametersROHM(),
        description="This dictionary contains the information about the parameters for the ROHM model.",
    )
    rohf: HomogenizedConductorFormulationparametersROHF = Field(
        default=HomogenizedConductorFormulationparametersROHF(),
        description="This dictionary contains the information about the parameters for the ROHF model.",
    )

class MultipoleSolve(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    coil_windings: Optional[MultipoleSolveCoilWindings] = Field(
        default=MultipoleSolveCoilWindings(),
        description="This dictionary contains the information pertaining the number of coils and electrical order necessary to generate the associated electrical circuit"
        )
    electromagnetics: MultipoleSolveElectromagnetics = Field(
        default=MultipoleSolveElectromagnetics(),
        description="This dictionary contains the solver information for the electromagnetic solution.",
    )
    thermal: MultipoleSolveThermal = Field(
        default=MultipoleSolveThermal(),
        description="This dictionary contains the solver information for the thermal solution.",
    )
    wedges: MultipoleSolveSpecificMaterial = Field(
        default=MultipoleSolveSpecificMaterial(),
        description="This dictionary contains the material information of wedges.",
    )
    collar: MultipoleSolveSpecificMaterial = Field(
        default=MultipoleSolveSpecificMaterial(),
        description="This dictionary contains the material information of the collar region.",
    )
    iron_yoke: MultipoleSolveSpecificMaterial = Field(
        default=MultipoleSolveSpecificMaterial(),
        description="This dictionary contains the material information of the iron yoke region.",
    )
    poles: MultipoleSolveSpecificMaterial = Field(
        default=MultipoleSolveSpecificMaterial(),
        description="This dictionary contains the material information of the pole region.",
    )
    noOfMPITasks: Optional[Union[bool, int]] = Field(
        default=False,
        title="No. of tasks for MPI parallel run of GetDP",
        description=(
            "If integer, GetDP will be run in parallel using MPI. This is only valid"
            " if MPI is installed on the system and an MPI-enabled GetDP is used." 
            " If False, GetDP will be run in serial without invoking mpiexec."
        ),
    )
    time_stepping: Optional[MultipoleSolveTransientCoupled] = Field(
        default=MultipoleSolveTransientCoupled(),
        description="This dictionary contains the information about the parameters for the transient solver.",
    )
    cable_homogenization: Optional[HomogenizedConductor]= Field(
        default=HomogenizedConductor(),
        description="This dictionary contains the information about the homogenized conductor properties.",
    )

class MultipoleThermalInsulationMesh(BaseModel):
    """
    Level 3: Class for FiQuS Multipole
    """
    global_size: float = Field(
        default=1e-4,
        description="It specifies the global size of the mesh for the insulation regions. It is enforced as a constant mesh field for surface insulation and by fixing the number of TSA layers for thin-shell approximation.",
    )
    TSA: Optional[MultipoleMeshThinShellApproximationParameters] = Field(
        default=MultipoleMeshThinShellApproximationParameters(),
        description="This dictionary contains the mesh information for thin-shells.",
    )

class MultipoleMesh_parent(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    create: bool = Field(
        default=True,
        description="It determines whether the mesh is built or not.",
    )
    conductors: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the conductor regions.",
    )
    wedges: Optional[MultipoleMeshTransfiniteOrField] = Field(
        default=MultipoleMeshTransfiniteOrField(),
        description="This dictionary contains the mesh information for the wedge regions.",
    )
    iron_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the iron yoke region.",
    )
    collar: Optional[MultipoleMeshThresholdCollar] = Field(
        default=MultipoleMeshThresholdCollar(),
        description="This dictionary contains the gmsh Field information for the collar region.",
    )
    poles: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the mesh information for the poles region.",
    )

class MultipoleMeshThermal(MultipoleMesh_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    reference: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="It determines whether the reference mesh is built or not. If True, an additional layer between the insulation and collar is meshed",
    )
    insulation: Optional[MultipoleThermalInsulationMesh] = Field(
        default=MultipoleThermalInsulationMesh(),
        description="This dictionary contains the mesh information for the insulation regions.",
    )
    isothermal_conductors: Optional[bool] = Field(
        default=False,
        description="It determines whether the conductors are considered isothermal or not using getDP constraints.",
    )
    isothermal_wedges: Optional[bool] = Field(
        default=False,
        description="It determines whether the wedges are considered isothermal or not using getDP Link constraints.",
    )

class MultipoleMeshElectromagnetics(MultipoleMesh_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    bore_field: Optional[MultipoleMeshThreshold] = Field(
        default=MultipoleMeshThreshold(),
        description="This dictionary contains the gmsh Field information for the bore region.",
    )

class MultipoleMesh(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    electromagnetics: MultipoleMeshElectromagnetics = Field(
        default=MultipoleMeshElectromagnetics(),
        description="This dictionary contains the mesh information for the electromagnetic solution.",
    )
    thermal: MultipoleMeshThermal = Field(
        default=MultipoleMeshThermal(),
        description="This dictionary contains the mesh information for the thermal solution.",
    )

class MultipoleGeometry_parent(BaseModel):
    create: bool = Field(
        default=True,
        description="It determines whether the geometry is built or not.",
    )
    with_wedges: Optional[bool] = Field(
        default=True,
        description="It determines whether the wedge regions are built or not.",
    )
    areas: Optional[List[Literal["iron_yoke", "collar", "poles"]]] = Field(
        default= [],
        description="List with areas to build."
    )

class MultipoleGeometryThermal(MultipoleGeometry_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    use_TSA: Optional[bool] = Field(
        default=False,
        description="It determines whether the insulation regions are explicitly built or modeled via thin-shell approximation.",
    )
    correct_block_coil_tsa_checkered_scheme: Optional[bool] = Field(
        default=False,
        description="There is a bug in the TSA naming scheme for block coils, this flag activates a simple (not clean) bug fix that will be replaced in a future version.",
    )
    use_TSA_new: Optional[bool] = Field(
        default=False,
        description="It determines whether the regions between collar and coils are modeled via thin-shell approximation.",
    )

class MultipoleGeometryElectromagnetics(MultipoleGeometry_parent):
    """
    Level 2: Class for FiQuS Multipole
    """
    symmetry: Optional[Literal["none", "xy", "x", "y"]] = Field(
        default='none',
        description="It determines the model regions to build according to the specified axis/axes.",
    )

class MultipoleGeometry(BaseModel):
    """
    Level 2: Class for FiQuS Multipole
    """
    geom_file_path: Optional[str] = Field(
        default=None,
        description="It contains the path to a .geom file. If null, the default .geom file produced by steam-sdk BuilderFiQuS will be used.",
    )
    plot_preview: Optional[bool] = Field(
        default=False,
        description="If true, it displays matplotlib figures of the magnet geometry with relevant information (e.g., conductor and block numbers).",
    )
    electromagnetics: MultipoleGeometryElectromagnetics = Field(
        default=MultipoleGeometryElectromagnetics(),
        description="This dictionary contains the geometry information for the electromagnetic solution.",
    )
    thermal: MultipoleGeometryThermal = Field(
        default=MultipoleGeometryThermal(),
        description="This dictionary contains the geometry information for the thermal solution.",
    )


class Multipole(BaseModel):
    """
    Level 1: Class for FiQuS Multipole
    """
    type: Literal["multipole"] = "multipole"
    geometry: MultipoleGeometry = Field(
        default=MultipoleGeometry(),
        description="This dictionary contains the geometry information.",
    )
    mesh: MultipoleMesh = Field(
        default=MultipoleMesh(),
        description="This dictionary contains the mesh information.",
    )
    solve: MultipoleSolve = Field(
        default=MultipoleSolve(),
        description="This dictionary contains the solution information.",
    )
    postproc: MultipolePostProc = Field(
        default=MultipolePostProc(),
        description="This dictionary contains the post-process information.",
    )


import timeit
import logging
from enum import Enum
import os
import subprocess
import re
import pandas as pd

import gmsh
import numpy as np

#from fiqus.data import RegionsModelFiQuS
from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel
# import fiqus.data.DataConductorACGeom as geom

from fiqus.pro_assemblers.ProAssembler import ASS_PRO

logger = logging.getLogger('FiQuS')

class Solve:
    def __init__(self, fdm, GetDP_path, geometry_folder, mesh_folder, verbose=True):
        self.fdm = fdm
        self.cacdm = fdm.magnet
        self.GetDP_path = GetDP_path
        self.solution_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.geometry_folder = geometry_folder
        self.mesh_folder = mesh_folder
        self.mesh_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.msh")
        self.pro_file = os.path.join(self.solution_folder, f"{self.magnet_name}.pro")
        self.regions_file = os.path.join(mesh_folder, f"{self.magnet_name}.regions")
        
        self.verbose = verbose
        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=fdm.run.verbosity_Gmsh)

        self.ass_pro = ASS_PRO(os.path.join(self.solution_folder, self.magnet_name))
        self.regions_model = FilesAndFolders.read_data_from_yaml(self.regions_file, RegionsModel)
        self.material_properties_model = None

        self.ed = {} # excitation dictionary

        gmsh.option.setNumber("General.Terminal", verbose)

    def assemble_pro(self):
        logger.info("Assembling .pro file")
        self.ass_pro.assemble_combined_pro(template = self.cacdm.solve.pro_template, rm = self.regions_model, dm = self.fdm, ed=self.ed, mp=self.material_properties_model)

    def read_excitation(self, inputs_folder_path):
        """
        Function for reading a CSV file for the 'piecewise' excitation case.

        :param inputs_folder_path: The full path to the folder with input files.
        :type inputs_folder_path: str
        """
        if self.cacdm.solve.source_parameters.source_type == 'piecewise' and self.cacdm.solve.source_parameters.piecewise.source_csv_file:
            input_file = os.path.join(inputs_folder_path, self.cacdm.solve.source_parameters.piecewise.source_csv_file)
            logger.info(f'Using excitation from file: {input_file}')
            df = pd.read_csv(input_file, delimiter=',', engine='python')
            excitation_time = df['time'].to_numpy(dtype='float').tolist()
            self.ed['time'] = excitation_time
            excitation_b = df['b'].to_numpy(dtype='float').tolist()
            self.ed['b'] = excitation_b
            excitation_I = df['I'].to_numpy(dtype='float').tolist()
            self.ed['I'] = excitation_I

    def run_getdp(self, solve = True, postOperation = True, gui = False):
        command = ["-v2", "-verbose", "3"]
        if solve: 
            command += ["-solve", "MagDyn", "-mat_mumps_icntl_14","100"] # icntl for mumps just by precaution
        command += ["-pos", "MagDyn"]

        logger.info(f"Running GetDP with command: {command}")
        startTime = timeit.default_timer()

        if self.cacdm.solve.general_parameters.noOfMPITasks:
            mpi_prefix = ["mpiexec", "-np", str(self.cacdm.solve.general_parameters.noOfMPITasks)]
        else:
            mpi_prefix = []

        getdpProcess = subprocess.Popen(mpi_prefix + [self.GetDP_path] + [self.pro_file] + command + ["-msh"] + [self.mesh_file], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        with getdpProcess.stdout:
            for line in iter(getdpProcess.stdout.readline, b""):
                line = line.decode("utf-8").rstrip()
                line = line.split("\r")[-1]
                if not "Test" in line:
                    if line.startswith("Info"):
                        parsedLine = re.sub(r"Info\s+:\s+", "", line)
                        logger.info(parsedLine)
                    elif line.startswith("Warning"):
                        parsedLine = re.sub(r"Warning\s+:\s+", "", line)
                        logger.warning(parsedLine)
                    elif line.startswith("Error"):
                        parsedLine = re.sub(r"Error\s+:\s+", "", line)
                        logger.error(parsedLine)
                        logger.error("Solving CAC failed.")
                        # raise Exception(parsedLine)
                    elif re.match("##", line):
                        logger.critical(line)
                    else:
                        logger.info(line)
        
        simulation_time = timeit.default_timer()-startTime
        
        if gui and ((postOperation and not solve) or (solve and postOperation)): # and self.cacdm.postproc.generate_pos_files
            # gmsh.option.setNumber("Geometry.Volumes", 1)
            # gmsh.option.setNumber("Geometry.Surfaces", 1)
            # gmsh.option.setNumber("Geometry.Curves", 1)
            # gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in posFiles:
                gmsh.open(os.path.join(self.solution_folder, posFile))
            self.gu.launch_interactive_GUI()
        else:
            if gmsh.isInitialized():
                gmsh.clear()
                gmsh.finalize()



from pydantic import BaseModel, Field
from typing import Dict, List, Union, Literal, Optional, Any

from fiqus.data.DataModelCommon import EnergyExtraction, CLIQ_Class, ESC_Class, E_CLIQ_Class, QuenchDetection, Circuit_Class, PowerSupplyClass, QuenchHeater
from fiqus.data.DataConductor import Conductor
from fiqus.data.DataFiQuSCCT import CCT
from fiqus.data.DataFiQuSMultipole import Multipole
from fiqus.data.DataFiQuSPancake3D import Pancake3D
from fiqus.data.DataFiQuSConductorAC_Strand import CACStrand
from fiqus.data.DataFiQuSHomogenizedConductor import HomogenizedConductor
from fiqus.data.DataFiQuSConductorAC_Rutherford import CACRutherford
from fiqus.data.DataFiQuSConductorAC_CC import CACCC

class RunFiQuS(BaseModel):
    """
    Class for FiQuS run
    """

    type: Literal[
        "start_from_yaml",
        "mesh_only",
        "geometry_only",
        "geometry_and_mesh",
        "pre_process_only",
        "mesh_and_solve_with_post_process_python",
        "solve_with_post_process_python",
        "solve_only",
        "post_process_getdp_only",
        "post_process_python_only",
        "post_process",
        "plot_python",
        "batch_post_process_python",
        "only_create_pro_file_from_mesh",
        "postprocess_veusz"
    ] = Field(
        default="start_from_yaml",
        title="Run Type of FiQuS",
        description="FiQuS allows you to run the model in different ways. The run type can be specified here. For example, you can just create the geometry and mesh or just solve the model with previous mesh, etc.",
    )
    geometry: Optional[Union[str, int]] = Field(
        default=None,
        title="Geometry Folder Key",
        description="This key will be appended to the geometry folder.",
    )
    mesh: Optional[Union[str, int]] = Field(
        default=None,
        title="Mesh Folder Key",
        description="This key will be appended to the mesh folder.",
    )
    solution: Optional[Union[str, int]] = Field(
        default=None,
        title="Solution Folder Key",
        description="This key will be appended to the solution folder.",
    )
    launch_gui: bool = Field(
        default=False,
        title="Launch GUI",
        description="If True, the GUI will be launched after the run.",
    )
    overwrite: bool = Field(
        default=False,
        title="Overwrite",
        description="If True, the existing folders will be overwritten, otherwise new folders will be created. NOTE: This setting has no effect for HTCondor runs.",
    )
    comments: str = Field(
        default="",
        title="Comments",
        description="Comments for the run. These comments will be saved in the run_log.csv file.",
    )
    verbosity_Gmsh: int = Field(
        default=5,
        title="verbosity_Gmsh",
        description="Level of information printed on the terminal and the message console (0: silent except for fatal errors, 1: +errors, 2: +warnings, 3: +direct, 4: +information, 5: +status, 99: +debug)",
    )
    verbosity_GetDP: int = Field(
        default=5,
        title="verbosity_GetDP",
        description="Level of information printed on the terminal and the message console. Higher number prints more, good options are 5 or 6.",
    )
    verbosity_FiQuS: bool = Field(
        default=True,
        title="verbosity_FiQuS",
        description="Level of information printed on the terminal and the message console by FiQuS. Only True of False for now.",
    )


class GeneralFiQuS(BaseModel):
    """
    Class for FiQuS general
    """

    magnet_name: Optional[str] = None


# class QuenchHeaters(BaseModel):
#     """
#     Level 3: Class for FiQuS
#     """

#     N_strips: Optional[int] = None  # set to 0 to avoid building quench heater thin shells
#     t_trigger: Optional[List[float]] = None
#     U0: Optional[List[float]] = None
#     C: Optional[List[float]] = None
#     R_warm: Optional[List[float]] = None
#     w: Optional[List[float]] = None
#     h: Optional[List[float]] = None
#     h_ins: List[List[float]] = []
#     type_ins: List[List[str]] = []
#     h_ground_ins: List[List[float]] = []
#     type_ground_ins: List[List[str]] = []
#     l: Optional[List[float]] = None
#     l_copper: Optional[List[float]] = None
#     l_stainless_steel: Optional[List[float]] = None
#     ids: Optional[List[int]] = None
#     turns: Optional[List[int]] = None
#     turns_sides: Optional[List[str]] = None



class QuenchProtection(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    energy_extraction: EnergyExtraction = EnergyExtraction()
    quench_heaters: QuenchHeater = QuenchHeater()
    cliq: CLIQ_Class = CLIQ_Class()
    esc: ESC_Class = ESC_Class()
    e_cliq: E_CLIQ_Class = E_CLIQ_Class()


class SolveDumpDataModel(BaseModel):
    solve: Any      # this is populated with magnet.solve section in MainFiQuS when dumping the yaml file for reference. The is no information on type in the solve section, so going for "Any" here
    circuit: Circuit_Class = Circuit_Class()
    power_supply: PowerSupplyClass = PowerSupplyClass()
    quench_protection: QuenchProtection = QuenchProtection()
    quench_detection: QuenchDetection = QuenchDetection()
    conductors: Dict[Optional[str], Conductor] = {}


class FDM(BaseModel):
    """
    Class for FiQuS
    """

    general: GeneralFiQuS = GeneralFiQuS()
    run: RunFiQuS = RunFiQuS()
    magnet: Union[Multipole, CCT, Pancake3D, CACStrand, HomogenizedConductor, CACRutherford, CACCC] = Field(
        default=Multipole(), discriminator="type"
    )
    circuit: Circuit_Class = Circuit_Class()
    power_supply: PowerSupplyClass = PowerSupplyClass()
    quench_protection: QuenchProtection = QuenchProtection()
    quench_detection: QuenchDetection = QuenchDetection()
    conductors: Dict[Optional[str], Conductor] = {}



from pydantic import BaseModel, Field
from typing import List, Literal, Optional, Union


# ============= GEOMETRY ============= #
# -- Input/Output settings -- #
class HomogenizedConductorIOsettingsLoad(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """

    load_from_yaml: Optional[bool] = Field(
        default=False,
        description="True to load the geometry from a YAML file, false to generate the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the YAML file from which to load the geometry.",
    )

class HomogenizedConductorIOsettingsSave(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """

    save_to_yaml: Optional[bool] = Field(
        default=False,
        description="True to save the geometry to a YAML-file, false to not save the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the output geometry YAML file.",
    )

class HomogenizedConductorGeometryIOsettings(BaseModel):
    """
    Level 2: Class for Input/Output settings for the cable geometry
    """

    load: HomogenizedConductorIOsettingsLoad = (
        HomogenizedConductorIOsettingsLoad()
    )
    save: HomogenizedConductorIOsettingsSave = (
        HomogenizedConductorIOsettingsSave()
    )

class Rectangle(BaseModel):
    """
    Level 2: Class for Input/Output settings for the cable geometry
    """
    center_position: Optional[List[float]] = Field(
        default=None, description="Center position in two dimensional plane (x, y)."
    )    
    width: Optional[float] = Field(
        default=None, description="Width of the region (m)."
    )
    height: Optional[float] = Field(
        default=None, description="Height of the region (m)."
    )

class Circle(BaseModel):
    """
    Level 2: Class for Input/Output settings for the cable geometry
    """
    center_position: Optional[List[float]] = Field(
        default=None, description="Center position in two dimensional plane (x, y)."
    )    
    radius: Optional[float] = Field(
        default=None, description="Radius of the circle (m)."
    )

# -- Strand geometry parameters -- #
class HomogenizedConductorGeometry(BaseModel):
    """
    Level 2: Class for strand geometry parameters
    """
    cables_definition: List[Rectangle] = Rectangle()
    excitation_coils: Optional[List[Rectangle]] = Field(default=[], description="List of excitation coils")
    air: Circle = Circle()
    air_form: Literal['circle'] = Field(
        default='circle', 
        description="Type of model geometry which will be generated. Supported options are only circle for now"
    )
    io_settings: HomogenizedConductorGeometryIOsettings = HomogenizedConductorGeometryIOsettings()

# ============= MESH ============= #

# -- Strand mesh settings -- #
class HomogenizedConductorMesh(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """

    scaling_global: Optional[float] = Field(
        default=1, description="Global scaling factor for mesh size."
    )
    air_boundary_mesh_size_ratio: Optional[float] = Field(
        default=1, description="Ratio within the air region from boundary to inner elements."
    )
    cable_mesh_size_ratio: Optional[float] = Field(
        default=1, description="Scaling factor within the cable regions."
    )


# ============= SOLVE ============= #
# -- General parameters -- #
class HomogenizedConductorSolveGeneralparameters(BaseModel):
    """
    Level 3: Class for general parameters
    """
    superconductor_linear: Optional[bool] = Field(default=False, description="For debugging: replace LTS by normal conductor.")

    noOfMPITasks: Optional[Union[bool, int]] = Field(
        default=False,
        title="No. of tasks for MPI parallel run of GetDP",
        description=(
            "If integer, GetDP will be run in parallel using MPI. This is only valid"
            " if MPI is installed on the system and an MPI-enabled GetDP is used." 
            " If False, GetDP will be run in serial without invoking mpiexec."
        ),
    )
    rho_cables: Optional[float] = Field(
        default=1,
        description='Resistance for cables when modelled as linear conductors (no current sharing with power law) [Ohm*m].'
    )
    strand_transposition_length: Optional[float] = Field(
        default=0.1,
        description='Transposition length of the strands in the Rutherford cable (m).'
    )
    n_strands: Optional[int] = Field(
        default=36,
        description='Number of strands in the cable (-).'
    )
    strand_filling_factor: Optional[float] = Field(
        default=0.8617,
        description='Filling factor of the strands in the rectangular cable envelope (-).'
    )


# -- Initial conditions -- #
class HomogenizedConductorSolveInitialconditions(BaseModel):
    """
    Level 3: Class for initial conditions
    """

    init_from_pos_file: bool = Field(
        default=False, description="This field is used to initialize the solution from a non-zero field solution stored in a .pos file."
    )
    pos_file_to_init_from: Optional[str] = Field(
        default=None,
        description="Name of .pos file for magnetic field (A/m) from which the solution should be initialized."
        " Should be in the Geometry_xxx/Mesh_xxx/ folder in which the Solution_xxx will be saved.",
    )


# -- Source parameters -- #
class SolveExcitationCoils(BaseModel):
    """
    Level 5: Class for superimposed DC field or current parameters for the sine source
    """
    enable: Optional[bool] = Field(default=False, description='Solve with excitation coils acting as sources.')


class SolveSineSourceSuperimposedDC(BaseModel):
    """
    Level 5: Class for superimposed DC field or current parameters for the sine source
    """
    field_magnitude: Optional[float] = Field(default=0.0, description="DC field magnitude (T) (direction along y-axis). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC field is used.")
    current_magnitude: Optional[float] = Field(default=0.0, description="DC current magnitude (A). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC current is used.")


class SolveSineSource(BaseModel):
    """
    Level 4: Class for Sine source parameters
    """
    frequency: Optional[float] = Field(default=None, description="Frequency of the sine source (Hz).")
    field_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine field (T).")
    current_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine current (A).")
    superimposed_DC: SolveSineSourceSuperimposedDC = SolveSineSourceSuperimposedDC()


class SolvePiecewiseSource(BaseModel):
    """
    Level 4: Class for piecewise (linear) source parameters
    """
    source_csv_file: Optional[str] = Field(default=None, description="File name for the from_file source type defining the time evolution of current and field (in-phase). Multipliers are used for each of them. The file should contain two columns: 'time' (s) and 'value' (field/current (T/A)), with these headers. If this field is set, times, applied_fields_relative and transport_currents_relative are ignored.")
    times: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear sources. Used only if source_csv_file is not set. Can be scaled by time_multiplier.")
    applied_fields_relative: Optional[List[float]] = Field(default=None, description="Applied fields relative to multiplier applied_field_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    transport_currents_relative: Optional[List[float]] = Field(default=None, description="Transport currents relative to multiplier transport_current_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    time_multiplier: Optional[float] = Field(default=None, description="Multiplier for the time values in times (scales the time values). Also used for the time values in the source_csv_file.")
    applied_field_multiplier: Optional[float] = Field(default=None, description="Multiplier for the applied fields in applied_fields_relative. Also used for the values in the source_csv_file.")
    transport_current_multiplier: Optional[float] = Field(default=None, description="Multiplier for the transport currents in transport_currents_relative. Also used for the values in the source_csv_file.")


class HomogenizedConductorSolveSourceparameters(BaseModel):
    """
    Level 3: Class for material properties
    """
    boundary_condition_type: Literal['Natural','Essential'] = Field(
        default='Natural',
        description="Type of boundary condition applied at the outer domain boundary.",
    )
    source_type: Literal['sine', 'piecewise'] = Field(
        default='sine',
        description="Time evolution of applied current and magnetic field. Supported options are: sine, sine_with_DC, piecewise_linear, from_list.",
    )
    parallel_resistor: Optional[Union[bool, float]] = Field(
        default=False,
        title="Resistor parallel to the cable(s)",
        description=(
            "If False, no parallel resistor and the current source directly and only feeds the cable."
            " If True, a resistor is placed in parallel with the cable, with a default resistance of 1 Ohm. If float (cannot be zero), this defines the value of the resistance."
            " If more than one cable is modelled, they are all connected in series (and carry the same current)." 
        ),
    )
    excitation_coils: SolveExcitationCoils = SolveExcitationCoils()

    sine: SolveSineSource = SolveSineSource()
    piecewise: SolvePiecewiseSource = SolvePiecewiseSource()
    field_angle: Optional[float] = Field(default=90, description="Angle of the source magnetic field, with respect to the x-axis (degrees).")
    cable_current_multipliers: Optional[List[float]] = Field(
        default = None,
        description = "Individual multipliers applied to the transport current imposed in each cable. factors are applied according to the cable declarations in the geometry section of the yaml."
    )   



# -- Numerical parameters -- #
class HomogenizedConductorNumericalparametersSine(BaseModel):
    """ 
    Level 4: Numerical parameters corresponding to the sine source
    """
    timesteps_per_period: Optional[float] = Field(default=None, description="Initial value for number of time steps (-) per period for the sine source. Determines the initial time step size.")
    number_of_periods_to_simulate: Optional[float] = Field(default=None, description="Number of periods (-) to simulate for the sine source.")


class HomogenizedConductorNumericalparametersPiecewise(BaseModel):
    """
    Level 4: Numerical parameters corresponding to the piecewise source
    """
    time_to_simulate: Optional[float] = Field(default=None, description="Total time to simulate (s). Used for the piecewise source.")
    timesteps_per_time_to_simulate: Optional[float] = Field(default=None, description="If variable_max_timestep is False. Number of time steps (-) per period for the piecewise source.")
    force_stepping_at_times_piecewise_linear: bool = Field(default=False, description="If True, time-stepping will contain exactly the time instants that are in the times_source_piecewise_linear list (to avoid truncation maximum applied field/current values).")

    variable_max_timestep: bool = Field(default=False, description="If False, the maximum time step is kept constant through the simulation. If True, it varies according to the piecewise definition.")
    times_max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear maximum time step.")
    max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Maximum time steps (s) at the times_max_timestep_piecewise_linear. Above the limits, linear extrapolation of the last two values.")


class HomogenizedConductorSolveNumericalparameters(BaseModel):
    """
    Level 3: Class for numerical parameters
    """

    sine: HomogenizedConductorNumericalparametersSine = HomogenizedConductorNumericalparametersSine()
    piecewise: HomogenizedConductorNumericalparametersPiecewise = HomogenizedConductorNumericalparametersPiecewise()


# -- FrequencyDomainSolver parameters -- #
class HomogenizedConductorSolveFrequencyDomainSweep(BaseModel):
    """ 
    Level 4: Class for the frequency sweep definition within a frequency domain solver.
    """
    run_sweep: Optional[bool] = Field(default=False, description='Enabling a frequency sweep.')

    start_frequency: Optional[float] = Field(default=1, description='Start frequency of the sweep in Hz.')
    end_frequency: Optional[float] = Field(default=100, description='End frequency of the sweep in Hz.')
    number_of_frequencies: Optional[int] = Field(default=3, description='Total number of frequencies in the sweep (logspaced)')


class HomogenizedConductorSolveFrequencyDomain(BaseModel):
    """
    Level 3: Class for frequency domain solver parameters
    """
    enable: Optional[bool] = Field(default=False, description='Enable frequency solver functionality in the solve step.')
    frequency_sweep: HomogenizedConductorSolveFrequencyDomainSweep = HomogenizedConductorSolveFrequencyDomainSweep()


# -- Formulation parameters -- #

class HomogenizedConductorFormulationparametersROHM(BaseModel):
    """
    Level 4: Class for ROHM model parameters
    """
    enable: Optional[bool] = Field(
        default=False, 
        description='Use ROHM to homogenize the magnetization hysteresis in the cables.'
    )
    parameter_csv_file: Optional[str] = Field(
        default=None,
        description='Name of the csv file containing the ROHM parameters within the inputs folder with expected row structure: [alpha,kappa,chi,gamma,lambda].'
    )
    weight_scaling: Optional[float] = Field(
        default=1.0,
        description='Downscaling factor (s<1.0) which is applied to all weights except the first, which is scaled up to compensate.'
    )
    tau_scaling: Optional[float] = Field(
        default=1.0,
        description='Scaling factor which is applied uniformly to all coupling time constants.'
    )

class HomogenizedConductorFormulationparametersDISCC(BaseModel):
    """
    Level 4: Class for DISCC model parameters
    """
    gamma_c: Optional[float] = Field(
        default=0.43,
        description='Main crossing scaling parameter (-) that quantifies crossing coupling due to field perpendicular to cable wide face.'
    )
    gamma_a: Optional[float] = Field(
        default=0.53,
        description='Main adjacent scaling parameter (-) that quantifies adjacent coupling due to field parallel to cable wide face.'
    )
    lambda_a: Optional[float] = Field(
        default=0.006,
        description='Mixing scaling parameter (-) that quantifies adjacent coupling due to field perpendicular to cable wide face.'
    )
    crossing_coupling_resistance: Optional[float] = Field(
        default=20e-6,
        description='Resistance (Ohm) of the contact between crossing strands.'
    )
    adjacent_coupling_resistance: Optional[float] = Field(
        default=10e-6,
        description='Resistance (Ohm) of the contact between adjacent strands over one periodicity length (strand twist pitch divided by the number of strands).'
    )
class HomogenizedConductorFormulationparametersROHF(BaseModel):
    """
    Level 4: Class for ROHF model parameters
    """
    enable: Optional[bool] = Field(
        default=False, 
        description='Use ROHF to homogenize the internal flux hysteresis in the cables.'
    )
    parameter_csv_file: Optional[str] = Field(
        default=None,
        description='Name of the csv file containing the ROHF parameters within the inputs folder with expected row structure: [alpha,kappa,tau].'
    )
class HomogenizedConductorFormulationparametersCS(BaseModel):
    """
    Level 4: Class for Current Sharing (CS) model parameters
    """
    superconductor_n_value: Optional[float] = Field(default=30, description="n value for the power law (-), used in current sharing law.")
    superconductor_Ic: Optional[float] = Field(default=350, description="Critical current of the strands (A) (e.g., typical value at T=1.9K and B=10T). Will be taken as a constant as in this model the field dependence is not included"
    " (the main purpose of the model is to verify the more efficient Homogenized Conductor model)."
    " Including field-dependence could be done but is not trivial because is mixes global and local quantities in this Rutherford model with strand discretized individually as stranded conductors.")
    matrix_resistance: Optional[float] = Field(default=6.536208e-04, description="Resistance of the matrix (per unit length) (Ohm/m) for the current sharing law. Kept constant in this model (for simplicity).")


class HomogenizedConductorFormulationparameters(BaseModel):
    """
    Level 3: Class for finite element formulation parameters
    """
    hphia: Optional[bool] = Field(default=False, description='Use hphia formulation.')



class HomogenizedConductorSolve(BaseModel):
    """
    Level 2: Class for FiQuS HomogenizedConductor solver settings
    """
    pro_template: Optional[Literal['HomogenizedConductor_template.pro']] = Field(
        default='HomogenizedConductor_template.pro',
        description="Name of the .pro template file."
    )
    general_parameters: HomogenizedConductorSolveGeneralparameters = (
        HomogenizedConductorSolveGeneralparameters()
    )
    formulation_parameters: HomogenizedConductorFormulationparameters = (
        HomogenizedConductorFormulationparameters()
    )
    discc: HomogenizedConductorFormulationparametersDISCC = (
        HomogenizedConductorFormulationparametersDISCC()
    )
    rohf: HomogenizedConductorFormulationparametersROHF = (
        HomogenizedConductorFormulationparametersROHF()
    )
    rohm: HomogenizedConductorFormulationparametersROHM = (
        HomogenizedConductorFormulationparametersROHM()
    )
    current_sharing: HomogenizedConductorFormulationparametersCS = (
        HomogenizedConductorFormulationparametersCS()
    )
    initial_conditions: HomogenizedConductorSolveInitialconditions = (
        HomogenizedConductorSolveInitialconditions()
    )
    source_parameters: HomogenizedConductorSolveSourceparameters = (
        HomogenizedConductorSolveSourceparameters()
    )
    numerical_parameters: HomogenizedConductorSolveNumericalparameters = (
        HomogenizedConductorSolveNumericalparameters()
    )
    frequency_domain_solver: HomogenizedConductorSolveFrequencyDomain = (
        HomogenizedConductorSolveFrequencyDomain()
    )


# ============= POSTPROC ============= #
class HomogenizedConductorFormulationparametersSampleLine(BaseModel):
    """
    Level 3: Class for sampling along a predefined line within the model
    """
    start_point: Optional[List[float]] = Field(
        default=None,
        description='Start point of the line in cartesian coordinates: [x,y,z].'
    )
    end_point: Optional[List[float]] = Field(
        default=None,
        description='End point of the line in cartesian coordinates: [x,y,z].'
    )
    samples: Optional[int] = Field(
        default=None,
        description='Integer number of evenly spaced sample points along the line including start and end point.'
    )

class HomogenizedConductorPostprocCleanup(BaseModel):
    """
    Level 3: Class for cleanup settings
    """
    remove_pre_file: bool = Field(
        default=False,
        description="Set True to remove the .pre-file after post-processing, to save disk space.",
    )
    remove_res_file: bool = Field(
        default=False,
        description="Set True to remove the .res-file after post-processing, to save disk space.",
    )
    remove_msh_file: bool = Field(
        default=False,
        description="Set True to remove the .msh-file after post-processing, to save disk space.",
    )


class HomogenizedConductorPostproc(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """
    generate_pos_files: bool = Field(
        default=True,
        description="Set True to generate .pos-files during post-processing",
    )
    output_folder: Optional[str] = Field(
        default=None,
        description="Batch post-processing creates a folder with the given name in the output directory, where all the plots are saved.",
    )
    generate_report: Optional[bool] = Field(
        default=False,
        description="Generates a PDF report including all postprocessing graphs. File is saved in the output_folder."
    )
    save_last_current_density: Optional[str] = Field(
        default=None,
        description="Saves the last current density field solution (out-of-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This can be for using the current density as an initial condition (but not implemented yet).",
    )
    save_last_magnetic_field: Optional[str] = Field(
        default=None,
        description="Saves the last magnetic field solution (in-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This is for using the magnetic field as an initial condition for another resolution.",
    )
    cleanup: HomogenizedConductorPostprocCleanup = HomogenizedConductorPostprocCleanup()
    sample_line:  HomogenizedConductorFormulationparametersSampleLine = (
        HomogenizedConductorFormulationparametersSampleLine()
    )


# ============= BASE ============= #
class HomogenizedConductor(BaseModel):
    """
    Level 1: Class for FiQuS ConductorAC
    """

    type: Literal["HomogenizedConductor"]
    geometry: HomogenizedConductorGeometry = HomogenizedConductorGeometry()
    mesh: HomogenizedConductorMesh = HomogenizedConductorMesh()
    solve: HomogenizedConductorSolve = HomogenizedConductorSolve()
    postproc: HomogenizedConductorPostproc = HomogenizedConductorPostproc()
    



import timeit
import json
import logging
import math
from enum import Enum
import operator
import itertools
import os
import pickle

import gmsh
import numpy as np

from fiqus.data import RegionsModelFiQuS
from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel

from fiqus.geom_generators.GeometryConductorAC_Rutherford import RutherfordCable
from fiqus.mesh_generators.MeshConductorAC_Strand_RutherfordCopy import Mesh
from abc import ABC, abstractmethod

occ = gmsh.model.occ
    
class CableMesh(Mesh):
    def __init__(self, fdm, verbose=True):
        super().__init__(fdm, verbose)
        self.geometry_class : RutherfordCable = None # Class from geometry generation step, used to store all information about tags corresponding to everything... To be changed later (probably)
    
    def load_geometry(self, geom_folder):
        """ Generating the geometry file also saves the geometry class as a .pkl file. This geometry class can be loaded to reference the different parts of the geometry for meshing."""
        geom_save_file = os.path.join(geom_folder, f'{self.magnet_name}.pkl')

        with open(geom_save_file, "rb") as geom_save_file: # Unnecessary to return geom instead of setting self.geometry_class
            geom = pickle.load(geom_save_file)
        return geom
    
    def generate_mesh(self, geom_folder):
        self.geometry_class : RutherfordCable = self.load_geometry(geom_folder)
        self.geometry_class.update_tags()
        self.geometry_class.add_physical_groups()

        # Mesh size field for strands:
        strand_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(strand_field, "SurfacesList", [strand.surface_tag for strand in self.geometry_class.strands])
        gmsh.model.mesh.field.setNumber(strand_field, "VIn", self.cacdm.mesh.strand_mesh_size_ratio * self.fdm.conductors[self.cacdm.solve.conductor_name].strand.diameter)

        # Mesh size field for coating:
        coating_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(coating_field, "SurfacesList", [self.geometry_class.coating.surface_tag])
        gmsh.model.mesh.field.setNumber(coating_field, "VIn", self.cacdm.mesh.coating_mesh_size_ratio * self.fdm.conductors[self.cacdm.solve.conductor_name].strand.diameter)

        # Mesh size field for air:
        air_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(air_field, "SurfacesList", [self.geometry_class.air[0].surface_tag])
        gmsh.model.mesh.field.setNumber(air_field, "VIn", self.cacdm.mesh.air_boundary_mesh_size_ratio * self.fdm.conductors[self.cacdm.solve.conductor_name].strand.diameter)

        # Mesh size field for excitation coils:
        coil_field = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.setNumbers(coil_field, "SurfacesList", [coil.surface_tag for coil in self.geometry_class.excitation_coils])
        gmsh.model.mesh.field.setNumber(coil_field, "VIn", 2*self.cacdm.mesh.coating_mesh_size_ratio * self.fdm.conductors[self.cacdm.solve.conductor_name].strand.diameter)

        total_meshSize_field = gmsh.model.mesh.field.add("Min")
        gmsh.model.mesh.field.setNumbers(total_meshSize_field, "FieldsList", [strand_field, coating_field, air_field, coil_field])
        
        gmsh.model.mesh.field.setAsBackgroundMesh(total_meshSize_field)

        gmsh.option.setNumber("Mesh.MeshSizeFactor", self.cacdm.mesh.scaling_global)
        gmsh.model.mesh.generate(2)

    def generate_cuts(self):
        """
            Computes the cuts for imposing global quantities (massive or stranded).
        """
        # We need:
        # 1) one cut for the coating,
        # 2) one cut per strand, which should be directly at the interface between the strand and the coating matrix,
        # 3) one cut per excitation coil region.

        # 1) Compute cut for the coating region and oriente it based on the surface orientation
        cable_outer_boundary = self.geometry_class.coating.physical_boundary_tag
        gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[cable_outer_boundary],dims=[1])
        air = self.geometry_class.air[0].physical_surface_tag
        coils = [coil.physical_surface_tag for coil in self.geometry_class.excitation_coils]
        gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[air]+coils,dims=[1])
        cuts = gmsh.model.mesh.computeHomology()
        gmsh.model.mesh.clearHomologyRequests()
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", str(cuts[0][1]))
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", str(cuts[1][1]))
        gmsh.plugin.run("HomologyPostProcessing")
        self.geometry_class.coating.cut_tag = cuts[1][1] + 1

        # 2) Compute cuts and get the oriented surface of each strand
        # Cuts are computed one by one for each strand.
        # iMPORTANT: the relative cohomology is computed.
        # We consider the boundaries of each strand, relative to all the other strand boundaries and inner air regions
        # so that we ensure to keep the support of the co-chains in interfaces between strands and coating matrix,
        # rather than between strands or adjacent with inner air regions.
        # NB: cuts could also be defined one by one, possibly going through other strands and inner air regions. (But it does not seem to work correctly now.)
        for strand in self.geometry_class.strands:
            gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[strand.physical_surface_tag], subdomainTags=[strand.physical_boundary_tag], dims=[2])
            gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[strand.physical_boundary_tag], subdomainTags=[other_strand.physical_boundary_tag for other_strand in self.geometry_class.strands if other_strand != strand]+[air], dims=[1])
        cuts = gmsh.model.mesh.computeHomology()
        gmsh.model.mesh.clearHomologyRequests()

        homology = cuts[1::2] # The homology results represent the surfaces of the strands
        cuts = cuts[::2] # The (randomly oriented) cuts on the boundary of the strands

        # Extract the tags from the homology and cuts
        homology = [h[1] for h in homology]
        cuts = [c[1] for c in cuts]

        # Format the homology and cuts for the plugin
        homology_formatted = ', '.join(map(str, homology)) 
        cuts_formatted = ', '.join(map(str, cuts))

        # Create an identity matrix for the transformation matrix
        N = len(self.geometry_class.strands)
        identity = '; '.join(', '.join('1' if i == j else '0' for j in range(N)) for i in range(N))

        # Next we get the boundaries of the oriented surfaces, which will also be oriented correctly
        # The result of this operation is the oriented boundaries of the strands
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", homology_formatted)
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", "")
        gmsh.plugin.setString("HomologyPostProcessing", "TransformationMatrix", identity)
        gmsh.plugin.setNumber("HomologyPostProcessing", "ApplyBoundaryOperatorToResults", 1)
        gmsh.plugin.run("HomologyPostProcessing")
        
        homology_oriented = [homology[-1] + i + 1 for i in range(N)]
        homology_oriented_formatted = ', '.join(map(str, homology_oriented))

        # Finally we get the cuts for each strand as some linear combination of the cuts, which allows inducing currents in each strand separately
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", homology_oriented_formatted)
        gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", cuts_formatted)
        gmsh.plugin.setNumber("HomologyPostProcessing", "ApplyBoundaryOperatorToResults", 0)
        gmsh.plugin.run("HomologyPostProcessing")

        for i, strand in enumerate(self.geometry_class.strands):
            # print(f"Strand {strand.physical_surface_name} cut tag: {homology_oriented[-1] + i + 1}")
            strand.cut_tag = int(homology_oriented[-1]) + i + 1

        # 3) Compute cuts for excitation coils (will be modelled as stranded conductors)
        strands = [strand.physical_surface_tag for strand in self.geometry_class.strands]
        coating = self.geometry_class.coating.physical_surface_tag
        for coil in self.geometry_class.excitation_coils:            
            coil_outer_boundary = coil.physical_boundary_tag
            gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[coil_outer_boundary],dims=[1])
            other_coils = [other_coil.physical_surface_tag for other_coil in self.geometry_class.excitation_coils if other_coil != coil]
            gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[air]+other_coils+strands+[coating],dims=[1])
            cuts = gmsh.model.mesh.computeHomology()
            gmsh.model.mesh.clearHomologyRequests()
            gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", str(cuts[0][1]))
            gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", str(cuts[1][1]))
            gmsh.plugin.run("HomologyPostProcessing")
            coil.cut_tag = cuts[1][1] + 1

    def generate_regions_file(self):
        """
            Generates the .regions file for the GetDP solver.
        """
        regions_model = RegionsModel()
        
        """ -- Initialize data -- """
        regions_model.powered["Strands"] = RegionsModelFiQuS.Powered()
        regions_model.powered["Strands"].vol.numbers = []
        regions_model.powered["Strands"].vol.names = []
        regions_model.powered["Strands"].surf.numbers = []
        regions_model.powered["Strands"].surf.names = []
        regions_model.powered["Strands"].cochain.numbers = []
        regions_model.powered["Strands"].cochain.names = []
        regions_model.powered["Strands"].curve.names = [] # Stores physical points at filament boundary (to fix phi=0)
        regions_model.powered["Strands"].curve.numbers = [] # Stores physical points at filament boundary (to fix phi=0)

        regions_model.powered["Coating"] = RegionsModelFiQuS.Powered()
        regions_model.powered["Coating"].vol.numbers = [self.geometry_class.coating.physical_surface_tag]
        regions_model.powered["Coating"].vol.names = [self.geometry_class.coating.physical_surface_name]
        regions_model.powered["Coating"].surf_out.numbers = [self.geometry_class.coating.physical_boundary_tag]
        regions_model.powered["Coating"].surf_out.names = [self.geometry_class.coating.physical_boundary_name]
        ## regions_model.powered["Coating"].surf_in.numbers = [ ]
        ## regions_model.powered["Coating"].surf_in.names = []
        regions_model.powered["Coating"].cochain.numbers = [self.geometry_class.coating.cut_tag]
        regions_model.powered["Coating"].cochain.names = [f"Cut: Coating"]
        regions_model.powered["Coating"].curve.names = ["EdgePoint: Coating"] # Stores physical points at filament boundary (to fix phi=0)
        regions_model.powered["Coating"].curve.numbers = [self.geometry_class.coating.physical_edge_point_tag] # Stores physical points at filament boundary (to fix phi=0)

        regions_model.air.point.names = []
        regions_model.air.point.numbers = []

        regions_model.powered["ExcitationCoils"] = RegionsModelFiQuS.Powered()
        regions_model.powered["ExcitationCoils"].vol.numbers = []
        regions_model.powered["ExcitationCoils"].vol.names = []
        regions_model.powered["ExcitationCoils"].surf.numbers = []
        regions_model.powered["ExcitationCoils"].surf.names = []
        regions_model.powered["ExcitationCoils"].cochain.numbers = []
        regions_model.powered["ExcitationCoils"].cochain.names = []

        for i, strand in enumerate(self.geometry_class.strands):
            regions_model.powered["Strands"].vol.numbers.append(strand.physical_surface_tag) # Surfaces in powered.vol
            regions_model.powered["Strands"].vol.names.append(strand.physical_surface_name)

            regions_model.powered["Strands"].surf.numbers.append(strand.physical_boundary_tag) # Boundaries in powered.surf
            regions_model.powered["Strands"].surf.names.append(strand.physical_boundary_name)

            regions_model.powered["Strands"].cochain.numbers.append(strand.cut_tag)
            regions_model.powered["Strands"].cochain.names.append(f"Cut: Strand {i}")

            # Add physical point at Strand boundary to fix phi=0
            regions_model.powered["Strands"].curve.names.append(f"EdgePoint: Strand {i}")
            regions_model.powered["Strands"].curve.numbers.append(strand.physical_edge_point_tag)

        
        regions_model.air.vol.number = self.geometry_class.air[0].physical_surface_tag
        regions_model.air.vol.name = self.geometry_class.air[0].physical_surface_name

        regions_model.air.surf.number = self.geometry_class.air[0].physical_boundary_tag
        regions_model.air.surf.name = self.geometry_class.air[0].physical_boundary_name

        for air_region in self.geometry_class.air[1:]:
            regions_model.air.point.names.append(air_region.physical_boundary_name)
            regions_model.air.point.numbers.append(air_region.physical_boundary_tag)
        
        for i, coil in enumerate(self.geometry_class.excitation_coils):
            regions_model.powered["ExcitationCoils"].vol.numbers.append(coil.physical_surface_tag) # Surfaces in powered.vol
            regions_model.powered["ExcitationCoils"].vol.names.append(coil.physical_surface_name)

            regions_model.powered["ExcitationCoils"].surf.numbers.append(coil.physical_boundary_tag) # Boundaries in powered.surf
            regions_model.powered["ExcitationCoils"].surf.names.append(coil.physical_boundary_name)

            regions_model.powered["ExcitationCoils"].cochain.numbers.append(coil.cut_tag)
            regions_model.powered["ExcitationCoils"].cochain.names.append(f"Cut: Coil {i}")

        # Add physical point at matrix boundary to fix phi=0
        ## regions_model.air.point.names = ["Point at matrix boundary"]
        ## regions_model.air.point.numbers = [self.geometry_class.matrix[-1].physicalEdgePointTag]

        FilesAndFolders.write_data_to_yaml(self.regions_file, regions_model.model_dump())


import timeit
import json
import logging
import math
from enum import Enum
import operator
import itertools
import os
import pickle
import subprocess
import re
import pandas as pd

import gmsh
import numpy as np

from fiqus.data import RegionsModelFiQuS
from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel
# import fiqus.data.DataConductorACGeom as geom

from fiqus.pro_assemblers.ProAssembler import ASS_PRO

logger = logging.getLogger('FiQuS')

class Solve:
    def __init__(self, fdm, GetDP_path, geometry_folder, mesh_folder, verbose=True):
        self.fdm = fdm
        self.cacdm = fdm.magnet
        self.GetDP_path = GetDP_path
        self.solution_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.geometry_folder = geometry_folder
        self.mesh_folder = mesh_folder
        self.mesh_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.msh")
        self.pro_file = os.path.join(self.solution_folder, f"{self.magnet_name}.pro")
        self.regions_file = os.path.join(mesh_folder, f"{self.magnet_name}.regions")
        
        self.verbose = verbose
        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=fdm.run.verbosity_Gmsh)

        self.ass_pro = ASS_PRO(os.path.join(self.solution_folder, self.magnet_name))
        self.regions_model = FilesAndFolders.read_data_from_yaml(self.regions_file, RegionsModel)
        self.material_properties_model = None

        self.ed = {} # excitation dictionary

        gmsh.option.setNumber("General.Terminal", verbose)

    def read_excitation(self, inputs_folder_path):
        """
        Function for reading csv for the 'from_file' excitation case
        :type inputs_folder_path: str
        """
        # if self.cacdm.solve.source_parameters.source_type == 'from_file':
        #     input_file = os.path.join(inputs_folder_path, self.cacdm.solve.source_parameters.source_csv_file)
        #     logger.info(f'Getting applied field and transport current from file: {input_file}')
        #     df = pd.read_csv(input_file, delimiter=',', engine='python')
        #     excitation_time = df['time'].to_numpy(dtype='float').tolist()
        #     self.ed['time'] = excitation_time
        #     excitation_b = df['b'].to_numpy(dtype='float').tolist()
        #     self.ed['b'] = excitation_b
        #     excitation_I = df['I'].to_numpy(dtype='float').tolist()
        #     self.ed['I'] = excitation_I

        if self.cacdm.solve.source_parameters.excitation_coils.enable and self.cacdm.solve.source_parameters.excitation_coils.source_csv_file:
            input_file = os.path.join(inputs_folder_path, self.cacdm.solve.source_parameters.excitation_coils.source_csv_file)
            logger.info(f'Getting excitation coils currents from file: {input_file}')
            df = pd.read_csv(input_file, delimiter=',', engine='python')
            
            if( len(df.columns) != len(self.cacdm.geometry.excitation_coils.centers)+1):
                logger.warning('Number of excitation coils in geometry ('+ str(len(self.cacdm.geometry.excitation_coils.centers))+') and input source file ('+str(len(df.columns))+') not compatible')

            excitation_time = df['time'].to_numpy(dtype='float').tolist()
            self.ed['time'] = excitation_time
            for i in range(1, len(df.columns)):
                Istr = 'I'+str(i)
                excitation_value = df[Istr].to_numpy(dtype='float').tolist()
                self.ed[Istr] = excitation_value

    # def get_material_properties(self, inputs_folder_path):
    #     """
    #     Function for reading material properties from the geometry YAML file.
    #     This reads the 'solution' section of the YAML file and stores it in the solution folder.
    #     This could also be a place to change the material properties in the future.
    #     """
    #     if self.cacdm.geometry.io_settings.load.load_from_yaml:
    #         input_yaml_file = os.path.join(inputs_folder_path, self.cacdm.geometry.io_settings.load.filename)
    #         Conductor_dm = FilesAndFolders.read_data_from_yaml(input_yaml_file, geom.Conductor)
    #         solution_parameters = Conductor_dm.Solution
    #         FilesAndFolders.write_data_to_yaml(os.path.join(self.solution_folder, "MaterialProperties.yaml"), solution_parameters.dict())
    #         self.material_properties_model = solution_parameters

        


    def assemble_pro(self):
        logger.info("Assembling .pro file")
        self.ass_pro.assemble_combined_pro(template = self.cacdm.solve.pro_template, rm = self.regions_model, dm = self.fdm, ed=self.ed, mp=self.material_properties_model)

    def run_getdp(self, solve = True, postOperation = True, gui = False):

        # f = self.cacdm.solve.source_parameters.frequency
        # bmax = self.cacdm.solve.source_parameters.applied_field_amplitude
        # Imax_ratio = self.cacdm.solve.source_parameters.ratio_of_max_imposed_current_amplitude

        command = ["-v2", "-verbose", "3"]
        if solve: 
            # command += ["-solve", "js_to_hs_2"] if not self.cacdm.solve.frequency_domain_solver.enable else ["-solve", "MagDyn_freq"] # for debugging purposes
            command += ["-solve", "MagDyn"] if not self.cacdm.solve.frequency_domain_solver.enable else ["-solve", "MagDyn_freq"] 
        # Solve_only seems to always call postproc, here we only save .pos-files if specified in input file
        # if (postOperation and not solve) or (solve and postOperation and self.cacdm.postproc.generate_pos_files): 
        # if self.cacdm.solve.formulation_parameters.dynamic_correction:
        #     command += " -pos MagDyn MagDyn_dynCorr"
        # else:
        command += ["-pos", "MagDyn"]

        logger.info(f"Running GetDP with command: {command}")
        startTime = timeit.default_timer()
        # subprocess.run(f"{self.GetDP_path} {self.pro_file} {command} -msh {self.mesh_file}")

        if self.cacdm.solve.general_parameters.noOfMPITasks:
            mpi_prefix = ["mpiexec", "-np", str(self.cacdm.solve.general_parameters.noOfMPITasks)]
        else:
            mpi_prefix = []

        getdpProcess = subprocess.Popen(mpi_prefix + [self.GetDP_path] + [self.pro_file] + command + ["-msh"] + [self.mesh_file], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        with getdpProcess.stdout:
            for line in iter(getdpProcess.stdout.readline, b""):
                line = line.decode("utf-8").rstrip()
                line = line.split("\r")[-1]
                if not "Test" in line:
                    if line.startswith("Info"):
                        parsedLine = re.sub(r"Info\s+:\s+", "", line)
                        logger.info(parsedLine)
                    elif line.startswith("Warning"):
                        parsedLine = re.sub(r"Warning\s+:\s+", "", line)
                        logger.warning(parsedLine)
                    elif line.startswith("Error"):
                        parsedLine = re.sub(r"Error\s+:\s+", "", line)
                        logger.error(parsedLine)
                        logger.error("Solving CAC failed.")
                        # raise Exception(parsedLine)
                    elif re.match("##", line):
                        logger.critical(line)
                    else:
                        logger.info(line)
        
        simulation_time = timeit.default_timer()-startTime
        # Save simulation time:
        if solve:
            logger.info(f"Solving Rutherford has finished in {round(simulation_time, 3)} seconds.")
            with open(self.solution_folder+f'\\txt_files\\simulation_time.txt', 'w') as file:
                file.write(str(simulation_time))


        if gui and ((postOperation and not solve) or (solve and postOperation and self.cacdm.postproc.generate_pos_files)):
            # gmsh.option.setNumber("Geometry.Volumes", 1)
            # gmsh.option.setNumber("Geometry.Surfaces", 1)
            # gmsh.option.setNumber("Geometry.Curves", 1)
            # gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in posFiles:
                gmsh.open(os.path.join(self.solution_folder, posFile))
            self.gu.launch_interactive_GUI()
        else:
            if gmsh.isInitialized():
                gmsh.clear()
                gmsh.finalize()

    def cleanup(self):
        """
            This funtion is used to remove .pre and .res files from the solution folder, as they may be large and not needed.
        """
        magnet_name = self.fdm.general.magnet_name
        cleanup = self.cacdm.postproc.cleanup

        if cleanup.remove_res_file:
            res_file_path = os.path.join(self.solution_folder, f"{magnet_name}.res")
            if os.path.exists(res_file_path):
                os.remove(res_file_path)
                logger.info(f"Removed {magnet_name}.res")
        
        if cleanup.remove_pre_file:
            pre_file_path = os.path.join(self.solution_folder, f"{magnet_name}.pre")
            if os.path.exists(pre_file_path):
                os.remove(pre_file_path)
                logger.info(f"Removed {magnet_name}.pre")

        if cleanup.remove_msh_file:
            msh_file_path = os.path.join(self.mesh_folder, f"{magnet_name}.msh")
            if os.path.exists(msh_file_path):
                os.remove(msh_file_path)
                logger.info(f"Removed {magnet_name}.msh")


import functools
import itertools
import operator
import os
import unittest
from typing import List

import numpy as np

from fiqus.data.DataFiQuS import FDM
from fiqus.utils.Utils import FilesAndFolders as FFs


class Paths:
    """
    Helper class used in FiQuS tests to get file and folder paths
    :param model_name: name of yaml input file (without .yaml)
    :param f_extension: file extension to apply to the folders, for example '.brep' or '.msh'
    :return: tuple: fdm, outputs_folder, input_folder, input_file, model_folder, model_file, reference_folder, reference_file
    """

    def __init__(self, model_name, f_extension=''):
        self.inputs_folder_name = '_inputs'
        self.outputs_folder_name = '_outputs'
        self.references_folder_name = '_references'

        self.test_outputs_folder = os.path.join(os.getcwd(), self.outputs_folder_name)
        self.inputs_folder = os.path.join(os.getcwd(), self.inputs_folder_name, model_name)
        self.model_folder = os.path.join(os.getcwd(), self.outputs_folder_name, model_name)
        self.references_folder = os.path.join(os.getcwd(), self.references_folder_name, model_name)

        self.input_file = os.path.join(self.inputs_folder, f'{model_name}.yaml')
        self.model_file = os.path.join(self.model_folder, f'{model_name}.{f_extension}')
        self.reference_file = os.path.join(self.references_folder, f'{model_name}.{f_extension}')
        self.reference_vi_file = os.path.join(self.references_folder, f'{model_name}.vi')
        self.reference_geo_yaml_file = os.path.join(self.references_folder, f'geometry.yaml')
        self.reference_mesh_yaml_file = os.path.join(self.references_folder, f'mesh.yaml')
        self.reference_regions_file = os.path.join(self.references_folder, f'{model_name}.regions')


def filecmp(filename1, filename2):
    """
    From: https://stackoverflow.com/questions/254350/in-python-is-there-a-concise-way-of-comparing-whether-the-contents-of-two-text
    Do the two files have exactly the same contents?
    """
    with open(filename1, "rb") as fp1, open(filename2, "rb") as fp2:
        print(f'The {filename1} size is: {os.fstat(fp1.fileno()).st_size} b')
        print(f'The {filename2} size is: {os.fstat(fp2.fileno()).st_size} b')
        if os.fstat(fp1.fileno()).st_size != os.fstat(fp2.fileno()).st_size:
            return False  # different sizes ∴ not equal

        # set up one 4k-reader for each file
        fp1_reader = functools.partial(fp1.read, 4096)
        fp2_reader = functools.partial(fp2.read, 4096)

        # pair each 4k-chunk from the two readers while they do not return '' (EOF)
        cmp_pairs = zip(iter(fp1_reader, b''), iter(fp2_reader, b''))

        # return True for all pairs that are not equal
        inequalities = itertools.starmap(operator.ne, cmp_pairs)
        ineqs = []
        for ineq in inequalities:
            ineqs.append(ineq)
        # voila; any() stops at first True value
        print(f'The file comp function gives: {not any(inequalities)}')
        return not any(inequalities)


def assert_two_parameters(true_value, test_value):
    """
     Some functions used in multiple test functions
        **Assert two parameters - accepts multiple types**
    """
    # TODO: improve robustness and readability
    test_case = unittest.TestCase()

    if isinstance(true_value, np.ndarray) or isinstance(true_value, list):
        if len(true_value) == 1:
            true_value = float(true_value)

    if isinstance(test_value, np.ndarray) or isinstance(test_value, list):
        if len(test_value) == 1:
            test_value = float(test_value)

    # Comparison
    if isinstance(test_value, np.ndarray) or isinstance(test_value, list):
        if np.array(true_value).ndim == 2:
            for i, test_row in enumerate(test_value):
                if isinstance(test_row[0], np.floating):
                    test_row = np.array(test_row).round(10)
                    true_value[i] = np.array(true_value[i]).round(10)

                test_case.assertListEqual(list(test_row), list(true_value[i]))
        else:
            if isinstance(test_value[0], np.floating):
                test_value = np.array(test_value).round(10)
                true_value = np.array(true_value).round(10)

            test_case.assertListEqual(list(test_value), list(true_value))
    else:
        test_case.assertEqual(test_value, true_value)


def assert_equal_readable_files(file1: str, file2: str, verbose: bool = False) -> None:
    """
    Assert that two readable (e.g. csv or cond) files are equal
    :param file1: full path to the first file
    :type file1: str
    :param file2: full path to the second file
    :type file2: str
    :param verbose: if ture more is printed to the output
    :type verbose: bool
    :return: Nothing, however it asserts using unittest testcase
    :rtype: None
    """
    test_case = unittest.TestCase()

    # Load files to compare
    with open(file1, "r") as f1, open(file2, "r") as f2:
        file1 = f1.readlines()
        file2 = f2.readlines()

    flag_equal = True
    for line in file2:
        if line not in file1:
            if verbose:
                print(f"Found line that appears in File2 but not in File1: {line}")
            flag_equal = False
    for line in file1:
        if line not in file2:
            if verbose:
                print(f"Found line that appears in File1 but not in File2: {line}")
            flag_equal = False

    # Display message and assert files have the same entries
    if flag_equal:
        if verbose:
            print('Files {} and {} have the same entries.'.format(file1, file2))
    test_case.assertEqual(flag_equal, True)


def compare_conductors_dicts(dict1, dict2, max_relative_error=1e-6):
    """
    Function that compares two conductor dicts and returns message list describing differences
    If there are no differences the list is empty
    :param dict1: reference dict to compare
    :param dict2: output dict to compare
    :param max_relative_error: maximum relative difference of values

    """

    the_same_flag = True
    for brick_id in dict1.keys():
        if dict1[brick_id] != dict2[brick_id]:
            pass
        else:
            for key, value1 in dict1[brick_id].items():
                value2 = dict2[brick_id][key]
                if key in ['SHAPE', 'DRIVELABEL']:
                    if value1 == value2:
                        pass
                    else:
                        the_same_flag = False
                else:
                    if abs(float(value1) - float(value2)) < max_relative_error * float(value2):
                        pass
                    elif float(value1) == float(value2):  # both zero
                        pass
                    else:
                        the_same_flag = False

    return the_same_flag



import unittest
from tests.utils.fiqus_test_classes import FiQuSSolverTests
from fiqus.data.DataFiQuS import FDM
import fiqus.data.DataFiQuSPancake3D as Pancake3D
import os
import platform
from pathlib import Path

linux_getdp_prefix_path = Path("/bin/cerngetdp/")
windows_getdp_prefix_path = Path("C:/cerngetdp/")


class TestSolvers(FiQuSSolverTests):
    def test_Pancake3D(self):
        """
        Checks if Pancake3D solvers work correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_PANCAKE3D") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"pancake3d/bin/getdp_{os.getenv('CERNGETDP_VERSION_PANCAKE3D')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"pancake3d/getdp_{os.getenv('CERNGETDP_VERSION_PANCAKE3D')}.exe")
        else:
            print("CERNGETDP_VERSION_PANCAKE3D is not set. Using default getdp path from data settings.")

        model_names = [
            "TEST_Pancake3D_REF",
            "TEST_Pancake3D_REFStructured",
            "TEST_Pancake3D_TSA",
            "TEST_Pancake3D_TSAStructured",
            "TEST_Pancake3D_TSAInsulating",
            "TEST_Pancake3D_TSAInsulatingJcVsLength",
            "TEST_Pancake3D_TSAInsulatingJcVsLength_thermalOnly",
        ]
        solve_types = ["electromagnetic", "weaklyCoupled", "stronglyCoupled", "thermal"]
        for model_name in model_names:
            for solve_type in solve_types:

                if solve_type == "thermal" and not model_name =="TEST_Pancake3D_TSAInsulatingJcVsLength_thermalOnly":
                    continue

                if solve_type != "thermal" and model_name =="TEST_Pancake3D_TSAInsulatingJcVsLength_thermalOnly":
                    continue

                with self.subTest(model_name=model_name, solve_type=solve_type):
                    data_model: FDM = self.get_data_model(model_name)

                    data_model.magnet.solve.type = solve_type

                    data_model.run.solution = solve_type

                    if solve_type in ["weaklyCoupled", "stronglyCoupled"]:
                        data_model.magnet.solve.quantitiesToBeSaved = [
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="magneticField",
                            ),
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="currentDensity",
                            ),
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="temperature",
                            ),
                        ]
                    elif solve_type == "electromagnetic":
                        data_model.magnet.solve.quantitiesToBeSaved = [
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="magneticField",
                            ),
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="currentDensity",
                            ),
                        ]
                    elif solve_type == "thermal":
                        data_model.magnet.solve.quantitiesToBeSaved = [
                            Pancake3D.Pancake3DSolveSaveQuantity(
                                quantity="temperature",
                            ),
                        ]

                    self.solve(data_model, model_name)

                    # Compare the pro files:
                    pro_file = self.get_path_to_generated_file(
                        data_model=data_model,
                        model_name=model_name,
                        file_extension="pro",
                    )
                    reference_pro_file = self.get_path_to_reference_file(
                        data_model=data_model,
                        model_name=model_name,
                        file_extension="pro",
                    )
                    self.compare_text_files(pro_file, reference_pro_file)

                    # Compare the results files:
                    if solve_type in ["electromagnetic", "weaklyCoupled", "stronglyCoupled"]:
                        pos_file = self.get_path_to_generated_file(
                            data_model=data_model,
                            model_name="MagneticField-DefaultFormat",
                            file_extension="pos",
                        )
                        reference_pos_file = self.get_path_to_reference_file(
                            data_model=data_model,
                            model_name="MagneticField-DefaultFormat",
                            file_extension="pos",
                        )
                        self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-3)

                    if solve_type in ["weaklyCoupled", "stronglyCoupled", "thermal"]:
                        pos_file = self.get_path_to_generated_file(
                            data_model=data_model,
                            model_name="Temperature-DefaultFormat",
                            file_extension="pos",
                        )
                        reference_pos_file = self.get_path_to_reference_file(
                            data_model=data_model,
                            model_name="Temperature-DefaultFormat",
                            file_extension="pos",
                        )
                        self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-3)

    def test_ConductorAC_Strand(self):
        """
        Checks if CACStrand solvers work correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_CAC_STRAND") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"cac_strand/bin/getdp_{os.getenv('CERNGETDP_VERSION_CAC_STRAND')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"cac_strand/getdp_{os.getenv('CERNGETDP_VERSION_CAC_STRAND')}.exe")
        else:
            print("CERNGETDP_VERSION_CAC_STRAND is not set. Using default getdp path from data settings.")


        model_names = [
            "TEST_CAC_Strand_hexFilaments",
            "TEST_CAC_Strand_adaptiveMesh",
            "TEST_CAC_wireInChannel",
        ]
        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model: FDM = self.get_data_model(model_name)

                self.solve(data_model, model_name)

                # Compare the pro files: 
                pro_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="pro",
                )
                reference_pro_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="pro",
                )
                # This makes no sense as long as the development on the Strand model pro-template is ongoing ...
                # Comparing the field solutions should ensure solver consistency without relying on the exact template structure. Skipping the pro-templates for now ~ AG
                #self.compare_text_files(pro_file, reference_pro_file)

                # Compare the magnetic flux density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="b_Omega",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="b_Omega",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-10)

                # Compare the current density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="j_OmegaC",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="j_OmegaC",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-10)

    def test_ConductorAC_Rutherford(self):
        """
        Checks if CAC_Rutherford solver works correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_CAC_RUTHERFORD") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"cac_rutherford/bin/getdp_{os.getenv('CERNGETDP_VERSION_CAC_RUTHERFORD')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"cac_rutherford/getdp_{os.getenv('CERNGETDP_VERSION_CAC_RUTHERFORD')}.exe")
        else:
            print("CERNGETDP_VERSION_CAC_RUTHERFORD is not set. Using default getdp path from data settings.")


        model_names = [
            "TEST_CAC_Rutherford",
        ]
        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model: FDM = self.get_data_model(model_name)

                self.solve(data_model, model_name)

                # Compare the current density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="jz",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="jz",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1E-10)

                # Compare the power loss files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="m",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="m",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1E-10)

    def test_ConductorAC_CC(self):
        """
        Checks if ConductorAC Coated Conductor solver works correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_CAC_CC") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"cac_cc/bin/getdp_{os.getenv('CERNGETDP_VERSION_CAC_CC')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"cac_cc/getdp_{os.getenv('CERNGETDP_VERSION_CAC_CC')}.exe")
        else:
            print("CERNGETDP_VERSION_CAC_CC is not set. Using default getdp path from data settings.")

        model_names = [
            "TEST_CAC_CC",
        ]
        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model: FDM = self.get_data_model(model_name)

                self.solve(data_model, model_name)

                # Compare the current density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="last_magnetic_field",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="last_magnetic_field",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-2, abs_tolerance=1E-10)

    def test_HomogenizedConductor(self):
        """
        Checks if HomogenizedConductor solver works correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_HOMOGENIZED_CONDUCTOR") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"homogenized_conductor/bin/getdp_{os.getenv('CERNGETDP_VERSION_HOMOGENIZED_CONDUCTOR')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"homogenized_conductor/getdp_{os.getenv('CERNGETDP_VERSION_HOMOGENIZED_CONDUCTOR')}.exe")
        else:
            print("CERNGETDP_VERSION_HOMOGENIZED_CONDUCTOR is not set. Using default getdp path from data settings.")

        model_names = [
            "TEST_HomogenizedConductor",
        ]
        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model: FDM = self.get_data_model(model_name)

                self.solve(data_model, model_name)

                # Compare the current density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="js",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="js",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1E-10)

                # Compare the power loss files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="p_tot",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="p_tot",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1E-10)

    def test_Multipole(self):
        """
        Checks if Multipole solvers work correctly by comparing the results to the
        reference results that were checked manually.
        """
        if os.getenv("CERNGETDP_VERSION_MULTIPOLE") is not None:
            os_name = platform.system()

            if os_name == "Linux":
                self.getdp_path = linux_getdp_prefix_path / Path(f"multipole/bin/getdp_{os.getenv('CERNGETDP_VERSION_MULTIPOLE')}")
            else:
                self.getdp_path = windows_getdp_prefix_path / Path(f"multipole/getdp_{os.getenv('CERNGETDP_VERSION_MULTIPOLE')}.exe")
        else:
            print("CERNGETDP_VERSION_MULTIPOLE is not set. Using default getdp path from data settings.")

        model_names = [
            "TEST_MULTIPOLE_MBH_1in1_TSA_withQH",
            "TEST_MULTIPOLE_MBH_1in1_TSA",
            "TEST_MULTIPOLE_MBH_1in1_REF",
            "TEST_MULTIPOLE_SMC_TSA_withQH",
            "TEST_MULTIPOLE_SMC_TSA",
            "TEST_MULTIPOLE_SMC_REF",
            "TEST_MULTIPOLE_4COND_TSA",
            "TEST_MULTIPOLE_FALCOND_C_TSA_COLLAR_POLE"
        ]
        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model: FDM = self.get_data_model(model_name)

                self.solve(data_model, model_name)

                # Compare the pro files: 
                pro_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="pro",
                )
                reference_pro_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="pro",
                )
                
                self.compare_text_files(pro_file, reference_pro_file, exclude_lines_keywords=["NameOfMesh", "Include"], exclude_first_n_lines=1)

                # Compare the magnetic flux density files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="b_Omega",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="b_Omega",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-2)

                # Compare the temperature files:
                pos_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="T_Omega_c",
                    file_extension="pos",
                )
                reference_pos_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="T_Omega_c",
                    file_extension="pos",
                )
                self.compare_pos_files(pos_file, reference_pos_file, rel_tolerance=1e-3, abs_tolerance=1e-2)

                # Compare the solve yaml files
                solve_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="solve",
                    file_extension="yaml",
                )
                reference_solve_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="solve",
                    file_extension="yaml",
                )
                self.compare_json_or_yaml_files(solve_file, reference_solve_file)


if __name__ == "__main__":
    unittest.main()



import os, cProfile

from fiqus.geom_generators.GeometryConductorAC_Strand import Geometry
from fiqus.mesh_generators.MeshConductorAC_Strand import Mesh, StrandMesh
from fiqus.getdp_runners.RunGetdpConductorAC_Strand import Solve
from fiqus.post_processors.PostProcessConductorAC import PostProcess, BatchPostProcess


class MainConductorAC_Strand:
    def __init__(self, fdm, inputs_folder_path='', outputs_folder_path='', verbose=True):
        """
        Main class for working with simulations for the Conductor AC model.
        :param fdm: FiQuS data model
        :param inputs_folder_path: full path to folder with input files
        :param verbose: if True, more info is printed in the console
        """
        self.verbose = verbose
        self.fdm = fdm
        self.inputs_folder_path = inputs_folder_path
        self.outputs_folder_path = outputs_folder_path
        self.GetDP_path = None
        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None
        self.model_file = None
        self.model_folder = None

    def generate_geometry(self, gui=False):
        """
        Generates the strand geometry.
        """
        os.chdir(self.geom_folder)
        g = Geometry(fdm=self.fdm, inputs_folder_path=self.inputs_folder_path, verbose=self.verbose)
        g.generate_strand_geometry(gui)

    def load_geometry(self, gui: bool = False):
        """
        Loads the previously generated geometry from the .brep file.
        """
        os.chdir(self.geom_folder)
        g = Geometry(fdm=self.fdm, inputs_folder_path=self.inputs_folder_path, verbose=self.verbose)
        g.load_conductor_geometry(gui)
        # self.model_file = g.model_file

    def pre_process(self, gui=False):
        pass

    def mesh(self, gui: bool = False):
        """
        Generates the mesh for the strand geometry.
        """
        os.chdir(self.mesh_folder)

        m = StrandMesh(fdm=self.fdm, verbose=self.verbose)
        m.generate_mesh(self.geom_folder)
        m.generate_cuts()
        m.generate_regions_file()
        m.save_mesh(gui)

        return {"test": 0}

    def load_mesh(self, gui=False):
        """
        Loads the previously generated mesh from the MSH file.
        """
        os.chdir(self.mesh_folder)
        m = Mesh(fdm=self.fdm, verbose=self.verbose)
        m.load_mesh(gui)

        # self.model_file = m.mesh_file

    def solve_and_postprocess_getdp(self, gui: bool = False):
        """
        Assembles the .pro-file from the template, then runs the simulation and the post-processing steps using GetDP.
        """
        os.chdir(self.solution_folder)

        s = Solve(self.fdm, self.GetDP_path, self.geom_folder, self.mesh_folder, self.verbose)
        s.read_excitation(inputs_folder_path=self.inputs_folder_path)
        s.get_solution_parameters_from_yaml(inputs_folder_path=self.inputs_folder_path)
        s.assemble_pro()
        s.run_getdp(solve=True, postOperation=True, gui=gui)
        s.cleanup()

    # def pre_process(self):
    #     os.chdir(self.solution_folder)

    #     s = Solve(self.fdm, self.GetDP_path, self.mesh_folder, self.verbose)

    def post_process_getdp(self, gui: bool = False):
        """
        Runs the post-processing steps trough GetDP.
        """
        os.chdir(self.solution_folder)

        s = Solve(self.fdm, self.GetDP_path, self.geom_folder, self.mesh_folder, self.verbose)
        s.read_excitation(inputs_folder_path=self.inputs_folder_path)
        s.get_solution_parameters_from_yaml(inputs_folder_path=self.inputs_folder_path)
        s.assemble_pro()
        s.run_getdp(solve=False, postOperation=True, gui=gui)

    #
    def post_process_python(self, gui: bool = False):
        # os.chdir(self.solution_folder)
        postProc = PostProcess(self.fdm, self.outputs_folder_path)

        if self.fdm.magnet.postproc.plot_instantaneous_power:
            postProc.instantaneous_loss()

        if self.fdm.magnet.postproc.plot_flux.show:
            postProc.internal_flux()

        postProc.plotter.show()

        return {'test': 0}

    def batch_post_process_python(self, gui: bool = False):
        """
        Runs batch post-processing steps using Python.
        Used for gathering, analysing, comparing and plotting data from multiple simulations.
        """
        BatchPostProc = BatchPostProcess(self.fdm, lossMap_gridData_folder=None, inputs_folder_path=self.inputs_folder_path, outputs_folder_path=self.outputs_folder_path)

        if self.fdm.magnet.postproc.batch_postproc.loss_map.produce_loss_map:
            # plotter.save_lossMap_gridData()
            # plotter.save_magnetization()
            BatchPostProc.plotter.create_lossMap()

        if self.fdm.magnet.postproc.batch_postproc.loss_map.cross_section.plot_cross_section:
            BatchPostProc.plotter.plot_lossMap_crossSection()

        if self.fdm.magnet.postproc.batch_postproc.loss_map.cross_section_sweep.animate_cross_section_sweep:
            BatchPostProc.plotter.animate_lossMap_crossSection()

        if self.fdm.magnet.postproc.batch_postproc.plot2d.produce_plot2d:
            BatchPostProc.plotter.plot2d()

        if self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.fit_rohf or self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.produce_error_map:
            BatchPostProc.rohf_on_grid()


